// aflpdata.h

#ifndef AFLPDATAHDR
#define AFLPDATAHDR

using namespace std;

#include "util.h"
#include "runsettings.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>

#define MISSING -9
#define ERROR   -99
#define END_OF_LINE -999
#define END_OF_FILE -9999

#define MAX_TAXA 200
#define MAX_MARKERS 2000
#define MAX_NAME 256

class AFLPData {
public:
  AFLPData(const string, const string);
  int getNumTaxa () const { return numTaxa; }
  int getNumMarkers () const { return numMarkers; }
  const string getTaxaName (int i) const { return taxaNames[i]; }
  int getAFLP (int i,int j) const { return aflp[i][j]; }
  double getMarkerLength (int i) const { return markerLength[i]; }
  void printSummary(ostream&);
  void makeDistMatrix(vector<vector<double> >&) const;
  void printDistMatrix(ostream&, vector<vector<double> >&, int) const;
  
private:
  int numTaxa;		  	  // Number of taxa.
  int numMarkers;		  // Number of markers.
  vector<string> taxaNames;        // taxaName[numTaxa]:
  				  //   names of the taxa
  vector<vector<int> > aflp;	  // aflp[numTaxa][numMarkers]:
  				  //   aflp[n][i] is the value of taxon n at
  				  //   site i of the original data file.
  vector<double> markerLength;            // markers[numMarkers].
  void readMarkerLength(ifstream &); 
  int convChar(char c);
  int readName(ifstream&,ostringstream&);
  int readAFLPFile(ifstream&);
  int readMarker(ifstream&,int&,char&);
};

#endif
