// graph.C

#include "util.h"
#include "graph.h"
#include "myfun.h"
#include "aflpdata.h"
#include <iostream>
#include <vector>

using namespace std;


Tree::Tree(const RunSettings& rs,const AFLPData& ad,Rand& rand) :
  numTaxa(ad.getNumTaxa()),
  numNodes(2*numTaxa-1), // rooted tree
  numEdges(2*numTaxa-2),
  numMarkers(ad.getNumMarkers()),
  nodes(numNodes),
  edges(numEdges)
{
  piC.resize(numMarkers);
  tProb2.resize(numMarkers);
  subLogProb16.resize(numMarkers);
  subLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    piC[i].resize(2);
    tProb2[i].resize(numEdges);
    subLogProb16[i].resize(numEdges);
    subLogProb18[i].resize(numEdges);
    for(int j=0; j<numEdges; j++){
      tProb2[i][j].resize(2);
      subLogProb16[i][j].resize(2);
      subLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	tProb2[i][j][k].resize(2);
	subLogProb16[i][j][k].resize(17);
	subLogProb18[i][j][k].resize(19);
      }
    }
  }
  tProb16.resize(numEdges);
  tProb18.resize(numEdges);
  for(int i=0; i<numEdges; i++){
    tProb16[i].resize(17);
    tProb18[i].resize(19);
    for(int j=0; j<17; j++)
      tProb16[i][j].resize(17);
    for(int j=0; j<19; j++)
      tProb18[i][j].resize(19);
  }

  tProb2X.resize(numMarkers);
  subLogProb16X.resize(numMarkers);
  subLogProb18X.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){

    tProb2X[i].resize(numEdges);
    subLogProb16X[i].resize(numEdges);
    subLogProb18X[i].resize(numEdges);
    for(int j=0; j<numEdges; j++){
      tProb2X[i][j].resize(2);
      subLogProb16X[i][j].resize(2);
      subLogProb18X[i][j].resize(2);
      for(int k=0; k<2; k++){
	tProb2X[i][j][k].resize(2);
	subLogProb16X[i][j][k].resize(17);
	subLogProb18X[i][j][k].resize(19);
      }
    }
  }
  tProb16X.resize(numEdges);
  tProb18X.resize(numEdges);
  for(int i=0; i<numEdges; i++){
    tProb16X[i].resize(17);
    tProb18X[i].resize(19);
    for(int j=0; j<17; j++)
      tProb16X[i][j].resize(17);
    for(int j=0; j<19; j++)
      tProb18X[i][j].resize(19);
  }

  for(int i=0;i<numTaxa;i++)
    nodes[i] = new Node(i,1);

  for(int i=numTaxa;i<numNodes;i++)
    nodes[i] = new Node(i,0);

  for(int i=0;i<numEdges;i++)
    edges[i] = new Edge(i);

  switch(rs.getInitialTreeType()){
  case RunSettings::GOOD: 
    createGoodTree(ad, rand);
    break;
  case RunSettings::RANDOM:
    createRandomTree(rs, rand); 
    break;
  }

}

void Tree::findSmallest(vector<int> &smallest, Node* node) const
{
  int num = node->getNumber();
  if(node->isLeaf())
    smallest[num] = num;
  else {
    Node *node1 = node->getNeighbor(1);
    findSmallest(smallest,node1);
    int n1 = smallest[node1->getNumber()];
    Node *node2 = node->getNeighbor(2);
    findSmallest(smallest,node2);
    int n2 = smallest[node2->getNumber()];
    smallest[num] = minimum(n1,n2);
  }
}

void Tree::printIntTop(ostream& f, vector<int> &smallest, Node* node) const
{
  if(node->isLeaf()) {
    f << node->getNumber()+1;
  }
  else {
    Node *node1 = node->getNeighbor(1);
    Node *node2 = node->getNeighbor(2);
    if(smallest[node1->getNumber()] < smallest[node2->getNumber()]) {
      f << "(";
      printIntTop(f,smallest,node1);
      f << ",";
      printIntTop(f,smallest,node2);
      f << ")";
    }
    else {
      f << "(";
      printIntTop(f,smallest,node2);
      f << ",";
      printIntTop(f,smallest,node1);
      f << ")";
    }
  }
}

void Tree::printTop(ostream& f) const
{
  vector<int> smallest(numNodes);
  findSmallest(smallest,nodes[root]);
  printIntTop(f,smallest,nodes[root]);
  f << ";" << endl;
}


void Tree::printIntBL(ostream& f, vector<int> &smallest, Node* node) const
{
  f << setiosflags(ios::showpoint | ios::fixed);
  if(node->isLeaf()) {
    f << node->getNumber()+1;
  }
  else {
    Node *node1 = node->getNeighbor(1);
    Node *node2 = node->getNeighbor(2);
    if(smallest[node1->getNumber()] < smallest[node2->getNumber()]) {
      f << "(";
      printIntBL(f,smallest,node1);
      f << ":" << setprecision(4) << setw(6) << node->getEdge(1)->getLength() << ",";
      printIntBL(f,smallest,node2);
      f << ":" << setprecision(4) << setw(6) << node->getEdge(2)->getLength() << ")";
    }
    else {
      f << "(";
      printIntBL(f,smallest,node2);
      f << ":" << setprecision(4) << setw(6) << node->getEdge(2)->getLength() << ",";
      printIntBL(f,smallest,node1);
      f << ":" << setprecision(4) << setw(6) << node->getEdge(1)->getLength() << ")";
    }
  }
}

void Tree::printBL(ostream& f) const
{
  vector<int> smallest(numNodes);
  findSmallest(smallest,nodes[root]);
  printIntBL(f,smallest,nodes[root]);
  f << ";" << endl;
}

void Edge::print(ostream& f) {
  f << "Edge(" << number << ")" << endl;
  f << "number = " << number << endl;
  f << "length = " << length << endl;
  f << "savedLength = " << savedLength << endl;
  f << "Nodes = (" << nodes[0]->getNumber() << "," << nodes[1]->getNumber() << ")" << endl;
}

void Edge::calculateTransitionMatricesM()
{
  int lowbd, upbd;
  double tempP;
  double p=3*(1-exp(-4*length/3))/4;

  int r=16;
  for(int j=0; j<=r; j++){
    for(int k=0; k<=r; k++){
      if(j-k<0){
	lowbd=0;
	tProb16[j][k] = choose(r-j, k-j) * pow(p,k-j) * pow(1-p, r-k) * pow(1-p/3, j);
	tempP=tProb16[j][k];
      }
      else{
	lowbd=j-k;
	tProb16[j][k] = pow(1-p, r-j) * choose(j,k) * pow(p/3,j-k) * pow(1-p/3, k);
	tempP=tProb16[j][k];
      }
      upbd = ((j<r-k) ? j:(r-k));  //min(j,r-k);
      for(int m=lowbd+1; m<=upbd; m++){
	tempP = tempP * (r-k-m+1) * (j-m+1) * pow(p,2) / ((k-j+m) * m * (1-p) * (3-p));
	tProb16[j][k] += tempP;
      }
    }
  }

  r=18;
  for(int j=0; j<=r; j++){
    for(int k=0; k<=r; k++){
      if(j-k<0){
	lowbd=0;
	tProb18[j][k] = choose(r-j, k-j) * pow(p,k-j) * pow(1-p, r-k) * pow(1-p/3, j);
	tempP=tProb18[j][k];
      }
      else{
	lowbd=j-k;
	tProb18[j][k] = pow(1-p, r-j) * choose(j,k) * pow(p/3,j-k) * pow(1-p/3, k);
	tempP=tProb18[j][k];
      }
      upbd = ((j<r-k) ? j:(r-k));  //min(j,r-k);
      for(int m=lowbd+1; m<=upbd; m++){
	tempP = tempP * (r-k-m+1) * (j-m+1) * pow(p,2) / ((k-j+m) * m * (1-p) * (3-p));
	tProb18[j][k] = tProb18[j][k] + tempP;
      }
    }
  }
}


void Edge::calculateTransitionMatricesC(double N)
{
  double q01 = ((N-4+1) * 12 / (pow(4.0, 4)-1) + ((N-6+1) * 18 / (pow(4.0,6)-1))) / 3;
  double Pi0 = pow(1-pow(.25,4), N-4+1) * pow(1-pow(.25,6), N-6+1);
  double temp=exp(-(q01*length)/(1-Pi0));
  tProb2[0][0] = Pi0+(1-Pi0)*temp;
  tProb2[0][1] = 1-tProb2[0][0];
  tProb2[1][0] = Pi0*(1-temp);
  tProb2[1][1] = 1-tProb2[1][0];   
}


void Node::calculateSubtreeLogLikelihoods(const AFLPData &ad, const RunSettings& rs, int k)
{
  edges[0]->calculateTransitionMatricesC(ad.getMarkerLength(k));

  double tuning;
  // Do not call on the root
  if(leaf) {
    for(int c=0;c<2;c++)
      for(int m=0;m<17;m++) {
	double tp16 = edges[0]->getTProb2(c,0) * edges[0]->getTProb16(m,0);
	subLogProb16[c][m] = (ad.getAFLP(number,k) == -9 ? 0 : (ad.getAFLP(number,k) == 1 ? log(tp16) : log(1 - tp16)));
      }
    for(int c=0;c<2;c++)
      for(int m=0;m<19;m++) {
	double tp18 = edges[0]->getTProb2(c,0) * edges[0]->getTProb18(m,0);
	subLogProb18[c][m] = (ad.getAFLP(number,k) == -9 ? 0 : (ad.getAFLP(number,k) == 1 ? log(tp18) : log(1 - tp18)));
      }
  }
  else {
    Edge *edge = edges[0];
    Node *node1 = getNeighbor(1);
    Node *node2 = getNeighbor(2);
    node1->calculateSubtreeLogLikelihoods(ad,rs,k);
    node2->calculateSubtreeLogLikelihoods(ad,rs,k);
    tuning = node1->getSubLogProb16(0,0)+ node2->getSubLogProb16(0,0);
    for(int c1=0;c1<2;c1++)
      for(int m1=0;m1<17;m1++) {
	double x = node1->getSubLogProb16(c1,m1)+ node2->getSubLogProb16(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    tuning *= -1;
    for(int c=0;c<2;c++)
      for(int m=0;m<17;m++) {
	double sum = 0.0;
	for(int c1=0;c1<2;c1++)
	  for(int m1=0;m1<17;m1++){
	    sum += edge->getTProb2(c,c1)*edge->getTProb16(m,m1)
	      * exp(node1->getSubLogProb16(c1,m1) + node2->getSubLogProb16(c1,m1) + tuning);
	  }
	subLogProb16[c][m] = log(sum)-tuning;
      }
    tuning = node1->getSubLogProb18(0,0)+ node2->getSubLogProb18(0,0);
    for(int c1=0;c1<2;c1++)
      for(int m1=0;m1<19;m1++) {
	double x = node1->getSubLogProb18(c1,m1)+ node2->getSubLogProb18(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    tuning *= -1;
    for(int c=0;c<2;c++)
      for(int m=0;m<19;m++) {
	double sum = 0.0;
	for(int c1=0;c1<2;c1++)
	  for(int m1=0;m1<19;m1++)
	    sum += edge->getTProb2(c,c1)*edge->getTProb18(m,m1)
	      * exp(node1->getSubLogProb18(c1,m1) + node2->getSubLogProb18(c1,m1) + tuning);
	subLogProb18[c][m] = log(sum) - tuning;
      }
  }
}


void Tree::calculatePiMs()
{
  for(int i=0; i<=16; i++)
    piM16[i] = choose(16,i) * pow(3.0, i) / pow(4.0, 16);
  for(int i=0; i<=18; i++)
    piM18[i] = choose(18,i) * pow(3.0, i) / pow(4.0, 18);
}

void Tree::calculatePiCs(int k, double N)
{
  piC[k][0] = pow((1-pow(.25, 4)),max(0,N-4+1)) * pow((1-pow(.25, 6)), max(0,N-6+1));
  piC[k][1] = 1-piC[k][0];
  if(piC[k][0]<0 || piC[k][1]<0){
    cerr << "error:" << piC[k][0] << "  " << piC[k][1] << "  " << N << endl;
    exit(1);
  }
}

double Tree::calculateLogLikelihood(const AFLPData &ad, const RunSettings& rs, int calPis)
{
  // Assumes that all edges point to parent nodes
  // \sum (pi_c \sum (pi_m * exp(logL+logr)))
  //  c           m
  logLikelihood = 0.0;
  Node *node1 = nodes[root]->getNeighbor(1);
  Node *node2 = nodes[root]->getNeighbor(2);
  int nodeNum1=node1->getNumber();
  int nodeNum2=node2->getNumber();
  double sum16, sum18, sum;
  double tuning;
  if(calPis)
    calculatePiMs();
  for(int e=0; e<numEdges; e++){
    edges[e]->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	setTProb16(e,i,j, edges[e]->getTProb16(i,j));
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	setTProb18(e,i,j, edges[e]->getTProb18(i,j));
  }

  for(int k=0;k<ad.getNumMarkers();k++) {
    if(calPis)
      calculatePiCs(k, ad.getMarkerLength(k));
    node1->calculateSubtreeLogLikelihoods(ad,rs,k);
    node2->calculateSubtreeLogLikelihoods(ad,rs,k);
    for(int n=0; n<numEdges; n++){   
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++)
	  setTProb2(k,n,i,j, edges[n]->getTProb2(i,j));
      for(int c=0; c<2; c++){
	for(int m=0; m<17; m++){
	  setSubLogProb16(k,n,c,m,nodes[n]->getSubLogProb16(c,m));
	}
	for(int m=0; m<19; m++)
	  setSubLogProb18(k,n,c,m,nodes[n]->getSubLogProb18(c,m));
      }
    }
    tuning = node1->getSubLogProb16(0,0)+ node2->getSubLogProb16(0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = node1->getSubLogProb16(c1,m1)+ node2->getSubLogProb16(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = node1->getSubLogProb18(c1,m1)+ node2->getSubLogProb18(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    sum16=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}

double Tree::calculateLogLikelihoodX(const AFLPData &ad, const RunSettings& rs)
{
  cout << setiosflags(ios::showpoint | ios::fixed);
  // Assumes that all edges point to parent nodes
  // \sum (pi_c \sum (pi_m * exp(logL+logr)))
  //  c           m
  logLikelihoodX = 0.0;
  Node *node1 = nodes[root]->getNeighbor(1);
  Node *node2 = nodes[root]->getNeighbor(2);
  double sum16, sum18, sum;
  for(int e=0; e<numEdges; e++){
    edges[e]->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	setTProb16X(e,i,j, edges[e]->getTProb16(i,j));
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	setTProb18X(e,i,j, edges[e]->getTProb18(i,j));
  }

  for(int k=0;k<ad.getNumMarkers();k++) {
    node1->calculateSubtreeLogLikelihoods(ad,rs,k);
    node2->calculateSubtreeLogLikelihoods(ad,rs,k);
    for(int n=0; n<numEdges; n++){   
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++)
	  setTProb2X(k,n,i,j, edges[n]->getTProb2(i,j));
      for(int c=0; c<2; c++){
	for(int m=0; m<17; m++){
	  setSubLogProb16X(k,n,c,m,nodes[n]->getSubLogProb16(c,m));
	}
	for(int m=0; m<19; m++)
	  setSubLogProb18X(k,n,c,m,nodes[n]->getSubLogProb18(c,m));
      }
    }

    sum16=0.0;
    double tuning;
    tuning = node1->getSubLogProb16(0,0)+ node2->getSubLogProb16(0,0);
    for(int c1=0;c1<2;c1++)
      for(int m1=0;m1<17;m1++) {
	double x = node1->getSubLogProb16(c1,m1)+ node2->getSubLogProb16(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    for(int c1=0;c1<2;c1++)
      for(int m1=0;m1<19;m1++) {
	double x = node1->getSubLogProb18(c1,m1)+ node2->getSubLogProb18(c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16X(k, node1->getNumber(), c,m) + getSubLogProb16X(k, node2->getNumber(), c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18X(k, node1->getNumber(), c,m) + getSubLogProb18X(k, node2->getNumber(), c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    /*
    cout << setiosflags(ios::showpoint | ios::scientific);
    cout << "*X*" << setw(4) << k << setw(15) << setprecision(6) << sum16 << setw(15) << setprecision(6) << sum18
	 << setw(15) << setprecision(6) << log((32*sum16+sum18)/33) - tuning << endl;
    cout.flush();
    cout << setiosflags(ios::showpoint | ios::fixed);
    */
    logLikelihoodX += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihoodX;
}


void Edge::setSavedLength()
{
  savedLength = length;  
}

void Edge::setBackLength()
{
  length = savedLength;
}

void Edge::setSavedProb2()
{
  for(int i=0; i<2; i++)
    for(int j=0; j<2; j++)
      savedTProb2[i][j] = tProb2[i][j];
}

void Edge::setBackProb2()
{
  for(int i=0; i<2; i++)
    for(int j=0; j<2; j++)
      tProb2[i][j] = savedTProb2[i][j];  
}


void Tree::calculateSubLogProb(Node* node, int lengChange, const AFLPData &ad, const RunSettings& rs)
{
  Edge* edge0=node->getEdge(0);
  int edgeNum = edge0->getNumber();
  int nodeNum = node->getNumber();

  if(lengChange) {
    edge0->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	tProb16[edgeNum][i][j] = edge0->getTProb16(i,j);
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	tProb18[edgeNum][i][j] = edge0->getTProb18(i,j);
  }

  for(int k=0; k<numMarkers; k++){
    if(lengChange) {
      edge0->calculateTransitionMatricesC(ad.getMarkerLength(k));
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++)
	  tProb2[k][edgeNum][i][j] = edge0->getTProb2(i,j);
    }

    double tuning;
    if(node->isLeaf()) {
      for(int c=0;c<2;c++)
	for(int m=0;m<17;m++) {
	  double tp16 = tProb2[k][edgeNum][c][0] * tProb16[edgeNum][m][0];
	  subLogProb16[k][nodeNum][c][m] = (ad.getAFLP(nodeNum,k) == -9 ? 0 : (ad.getAFLP(nodeNum,k) == 1 ? log(tp16) : log(1 - tp16)));
      }
      for(int c=0;c<2;c++)
	for(int m=0;m<19;m++) {
	  double tp18 = tProb2[k][edgeNum][c][0] * tProb18[edgeNum][m][0];
	  subLogProb18[k][nodeNum][c][m] = (ad.getAFLP(nodeNum,k) == -9 ? 0 : (ad.getAFLP(nodeNum,k) == 1 ? log(tp18) : log(1 - tp18)));
      }
    }
    else {
      int nodeNum1 = node->getNeighbor(1)->getNumber();
      int nodeNum2 = node->getNeighbor(2)->getNumber();
      tuning = subLogProb16[k][nodeNum1][0][0]+ subLogProb16[k][nodeNum2][0][0];
      for(int c1=0;c1<2;c1++)
	for(int m1=0;m1<17;m1++) {
	  double x = subLogProb16[k][nodeNum1][c1][m1]+ subLogProb16[k][nodeNum2][c1][m1];
	  if(x > tuning)
	    tuning = x;
	}
      tuning *= -1;
      for(int c=0;c<2;c++)
	for(int m=0;m<17;m++) {
	  double sum = 0.0;
	  for(int c1=0;c1<2;c1++)
	    for(int m1=0;m1<17;m1++){
	      sum += tProb2[k][edgeNum][c][c1] * tProb16[edgeNum][m][m1]
		* exp(subLogProb16[k][nodeNum1][c1][m1] + subLogProb16[k][nodeNum2][c1][m1] + tuning);
	  }
	subLogProb16[k][nodeNum][c][m] = log(sum)-tuning;
      }
      tuning = subLogProb18[k][nodeNum1][0][0]+ subLogProb18[k][nodeNum2][0][0];
      for(int c1=0;c1<2;c1++)
	for(int m1=0;m1<19;m1++) {
	  double x = subLogProb18[k][nodeNum1][c1][m1]+ subLogProb18[k][nodeNum2][c1][m1];
	  if(x > tuning)
	    tuning = x;
	}
      tuning *= -1;
      for(int c=0;c<2;c++)
	for(int m=0;m<19;m++) {
	  double sum = 0.0;
	  for(int c1=0;c1<2;c1++)
	    for(int m1=0;m1<19;m1++)
	      sum += tProb2[k][edgeNum][c][c1] * tProb18[edgeNum][m][m1]
		* exp(subLogProb18[k][nodeNum1][c1][m1] + subLogProb18[k][nodeNum2][c1][m1] + tuning);
	  subLogProb18[k][nodeNum][c][m] = log(sum) - tuning;
	}
    }
  }
}


void Tree::setSavedProbs0()
{
  savedTProb2.resize(numMarkers);
  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(numEdges);
    savedSubLogProb16[i].resize(numEdges);
    savedSubLogProb18[i].resize(numEdges);
    for(int j=0; j<numEdges; j++){
      savedTProb2[i][j].resize(2);
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedTProb2[i][j][k].resize(2);
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<2; m++)
	  savedTProb2[i][j][k][m] = tProb2[i][j][k][m];
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][j][k][m];
      }
    }
  }
  savedTProb16.resize(numEdges);
  savedTProb18.resize(numEdges);
  for(int i=0; i<numEdges; i++){
    savedTProb16[i].resize(17);
    savedTProb18[i].resize(19);
    for(int j=0; j<17; j++){
      savedTProb16[i][j].resize(17);
      for(int k=0; k<17; k++)
	savedTProb16[i][j][k] = tProb16[i][j][k];
    }
    for(int j=0; j<19; j++){
      savedTProb18[i][j].resize(19);
      for(int k=0; k<19; k++)
	savedTProb18[i][j][k] = tProb18[i][j][k];
    }
  }
}


void Tree::setBackProbs0()
{
  for(int i=0; i<numMarkers; i++)
    for(int j=0; j<numEdges; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<2; m++)
	  tProb2[i][j][k][m] = savedTProb2[i][j][k][m];
	for(int m=0; m<17; m++)
	  subLogProb16[i][j][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][j][k][m] = savedSubLogProb18[i][j][k][m];
      }
  for(int i=0; i<numEdges; i++){
    for(int j=0; j<17; j++)
      for(int k=0; k<17; k++)
	tProb16[i][j][k] = savedTProb16[i][j][k];
    for(int j=0; j<19; j++)
      for(int k=0; k<19; k++)
	tProb18[i][j][k] = savedTProb18[i][j][k];
  }
}


void Tree::setSavedProbs1(int edgeChosen)
  // tProb2[edgeChosen],tProb16[edgeChosen],tProb18[edgeChosen]
  // a list of subLogProb16, subLogProb18
{
  savedTProb2.resize(numMarkers);
  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(1);
    savedTProb2[i][0].resize(2);
    for(int j=0; j<2; j++){
      savedTProb2[i][0][j].resize(2);
      for(int k=0; k<2; k++)
	savedTProb2[i][0][j][k] = tProb2[i][edgeChosen][j][k];
    }
    savedSubLogProb16[i].resize(numNodesToRoot);
    savedSubLogProb18[i].resize(numNodesToRoot);
    for(int j=0; j<numNodesToRoot; j++){
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }
  savedTProb16.resize(1);
  savedTProb18.resize(1);
  savedTProb16[0].resize(17);
  savedTProb18[0].resize(19);
  for(int i=0; i<17; i++){
    savedTProb16[0][i].resize(17);
    for(int j=0; j<17; j++)
      savedTProb16[0][i][j] = tProb16[edgeChosen][i][j];
  }
  for(int i=0; i<19; i++){
    savedTProb18[0][i].resize(19);
    for(int j=0; j<19; j++)
      savedTProb18[0][i][j] = tProb18[edgeChosen][i][j];
  }
}

void Tree::setBackProbs1(int edgeChosen)
{
  for(int k=0; k<numMarkers; k++){
    for(int i=0; i<2; i++)
      for(int j=0; j<2; j++)
	tProb2[k][edgeChosen][i][j] = savedTProb2[k][0][i][j];
    for(int n=0; n<numNodesToRoot; n++)
      for(int i=0; i<2; i++){
	for(int j=0; j<17; j++)
	  subLogProb16[k][indNodesToRoot[n]][i][j] = savedSubLogProb16[k][n][i][j];
	for(int j=0; j<19; j++)
	  subLogProb18[k][indNodesToRoot[n]][i][j] = savedSubLogProb18[k][n][i][j];
      }
  }
  for(int i=0; i<17; i++)
    for(int j=0; j<17; j++)
      tProb16[edgeChosen][i][j] = savedTProb16[0][i][j];
  for(int i=0; i<19; i++)
    for(int j=0; j<19; j++)
      tProb18[edgeChosen][i][j] = savedTProb18[0][i][j];
}


double Tree::calculateNewLogLikelihood1(const AFLPData &ad, const RunSettings& rs)
{
  logLikelihood = 0.0;

  calculateSubLogProb(nodes[indNodesToRoot[0]],1, ad, rs);
  for(int i=1; i<numNodesToRoot; i++)
    calculateSubLogProb(nodes[indNodesToRoot[i]],0, ad, rs);

  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  if(indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2){
    cerr << " error: indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2" << endl;
    exit(1);
  }
  
  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = getSubLogProb16(k,nodeNum1,0,0)+ getSubLogProb16(k, nodeNum2,0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = getSubLogProb16(k, nodeNum1,c1,m1)+ getSubLogProb16(k, nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = getSubLogProb18(k, nodeNum1,c1,m1)+ getSubLogProb18(k, nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}


void Tree::setSavedProbs2()
{
  savedTProb2.resize(numMarkers);
  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(numNodesToRoot);
    savedSubLogProb16[i].resize(numNodesToRoot);
    savedSubLogProb18[i].resize(numNodesToRoot);
    for(int j=0; j<numNodesToRoot; j++){
      savedTProb2[i][j].resize(2);
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedTProb2[i][j][k].resize(2);
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<2; m++)
	  savedTProb2[i][j][k][m] = tProb2[i][indEdgesToRoot[j]][k][m];
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }
  savedTProb16.resize(numNodesToRoot);
  savedTProb18.resize(numNodesToRoot);
  for(int i=0; i<numNodesToRoot; i++){
    savedTProb16[i].resize(17);
    savedTProb18[i].resize(19);
    for(int j=0; j<17; j++){
      savedTProb16[i][j].resize(17);
      for(int k=0; k<17; k++)
	savedTProb16[i][j][k] = tProb16[indEdgesToRoot[i]][j][k];
    }
    for(int j=0; j<19; j++){
      savedTProb18[i][j].resize(19);
      for(int k=0; k<19; k++)
	savedTProb18[i][j][k] = tProb18[indEdgesToRoot[i]][j][k];
    }
  }
}


void Tree::setBackProbs2()
{
  for(int i=0; i<numMarkers; i++)
    for(int j=0; j<numNodesToRoot; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<2; m++)
	  tProb2[i][indEdgesToRoot[j]][k][m] = savedTProb2[i][j][k][m];
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesToRoot[j]][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesToRoot[j]][k][m] = savedSubLogProb18[i][j][k][m];
      }
  for(int i=0; i<numNodesToRoot; i++){
    for(int j=0; j<17; j++)
      for(int k=0; k<17; k++)
	tProb16[indEdgesToRoot[i]][j][k] = savedTProb16[i][j][k];
    for(int j=0; j<19; j++)
      for(int k=0; k<19; k++)
	tProb18[indEdgesToRoot[i]][j][k] = savedTProb18[i][j][k];
  }
}


double Tree::calculateNewLogLikelihood2(const AFLPData &ad, const RunSettings& rs)
{
  logLikelihood = 0.0;
  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = subLogProb16[k][nodeNum1][0][0]+ subLogProb16[k][nodeNum2][0][0];
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = subLogProb16[k][nodeNum1][c1][m1]+ subLogProb16[k][nodeNum2][c1][m1];
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = subLogProb18[k][nodeNum1][c1][m1]+ subLogProb18[k][nodeNum2][c1][m1];
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}


void Tree::setSavedProbs3()
{
  savedTProb2.resize(numMarkers);
  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(numEdgesToRoot);
    for(int j=0; j<numEdgesToRoot; j++){
      savedTProb2[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedTProb2[i][j][k].resize(2);
	for(int m=0; m<2; m++)
	  savedTProb2[i][j][k][m] = tProb2[i][indEdgesToRoot[j]][k][m];
      }
    }
    savedSubLogProb16[i].resize(numNodesToRoot);
    savedSubLogProb18[i].resize(numNodesToRoot);
    for(int j=0; j<numNodesToRoot; j++){
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }
  savedTProb16.resize(numEdgesToRoot);
  savedTProb18.resize(numEdgesToRoot);
  for(int i=0; i<numEdgesToRoot; i++){
    savedTProb16[i].resize(17);
    savedTProb18[i].resize(19);
    for(int j=0; j<17; j++){
      savedTProb16[i][j].resize(17);
      for(int k=0; k<17; k++)
	savedTProb16[i][j][k] = tProb16[indEdgesToRoot[i]][j][k];
    }
    for(int j=0; j<19; j++){
      savedTProb18[i][j].resize(19);
      for(int k=0; k<19; k++)
	savedTProb18[i][j][k] = tProb18[indEdgesToRoot[i]][j][k];
    }
  }
}


void Tree::setBackProbs3()
{
  for(int i=0; i<numMarkers; i++){
    for(int j=0; j<numNodesToRoot; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesToRoot[j]][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesToRoot[j]][k][m] = savedSubLogProb18[i][j][k][m];
      }
    for(int j=0; j<numEdgesToRoot; j++)
      for(int k=0; k<2; k++)
	for(int m=0; m<2; m++)
	  tProb2[i][indEdgesToRoot[j]][k][m] = savedTProb2[i][j][k][m];
  }
  for(int i=0; i<numEdgesToRoot; i++){
    for(int j=0; j<17; j++)
      for(int k=0; k<17; k++)
	tProb16[indEdgesToRoot[i]][j][k] = savedTProb16[i][j][k];
    for(int j=0; j<19; j++)
      for(int k=0; k<19; k++)
	tProb18[indEdgesToRoot[i]][j][k] = savedTProb18[i][j][k];
  }
}



double Tree::calculateNewLogLikelihood3(const AFLPData &ad, const RunSettings& rs)
{
  logLikelihood = 0.0;
  for(int n=0; n<numEdgesToRoot; n++) {
    int edgeNum = indEdgesToRoot[n];
    Edge* edge0 = edges[edgeNum];
    edge0->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	setTProb16(edgeNum,i,j, edge0->getTProb16(i,j));
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	setTProb18(edgeNum,i,j, edge0->getTProb18(i,j));
    for(int k=0; k<numMarkers; k++){
      edge0->calculateTransitionMatricesC(ad.getMarkerLength(k));
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++){
	  setTProb2(k,edgeNum,i,j, edge0->getTProb2(i,j));
	}
    }
  }
  for(int i=0; i<numNodesToRoot; i++)
    calculateSubLogProb(nodes[indNodesToRoot[i]],0, ad, rs);

  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  if(indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2){
    cerr << " error: indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2" << endl;
    exit(1);
  }
  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = getSubLogProb16(k, nodeNum1,0,0)+ getSubLogProb16(k,nodeNum2,0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = getSubLogProb16(k, nodeNum1,c1,m1)+ getSubLogProb16(k,nodeNum2,c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = getSubLogProb18(k, nodeNum1, c1,m1)+ getSubLogProb18(k,nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}


void Tree::setSavedValues4()
{
  savedNodes.resize(numNodesInPath);
  savedEdges.resize(numEdgesInPath);
  for(int i=0; i<numEdgesInPath; i++){
    savedNodes[i]=*nodes[indNodesInPath[i]];
    savedEdges[i]=*edges[indEdgesInPath[i]];
  }
  savedNodes[numEdgesInPath]=*nodes[indNodesInPath[numEdgesInPath]];
  int numNodesForSubLog=numNodesInPath2+numNodesToRoot;

  savedTProb2.resize(numMarkers);
  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(numEdgesInPath);
    for(int j=0; j<numEdgesInPath; j++){
      savedTProb2[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedTProb2[i][j][k].resize(2);
	for(int m=0; m<2; m++)
	  savedTProb2[i][j][k][m] = tProb2[i][indEdgesInPath[j]][k][m];
      }
    }
    savedSubLogProb16[i].resize(numNodesForSubLog);
    savedSubLogProb18[i].resize(numNodesForSubLog);
    for(int j=0; j<numNodesForSubLog; j++){
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
      }
    }
    for(int j=0; j<numNodesInPath2; j++){
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++){
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesInPath[j]][k][m];
	}
	for(int m=0; m<19; m++){
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesInPath[j]][k][m];
	}
      }
    }
    for(int j=0; j<numNodesToRoot; j++){
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++){
	  savedSubLogProb16[i][numNodesInPath2+j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	}
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][numNodesInPath2+j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }
  savedTProb16.resize(numEdgesInPath);
  savedTProb18.resize(numEdgesInPath);
  for(int i=0; i<numEdgesInPath; i++){
    savedTProb16[i].resize(17);
    savedTProb18[i].resize(19);
    for(int j=0; j<17; j++){
      savedTProb16[i][j].resize(17);
      for(int k=0; k<17; k++)
	savedTProb16[i][j][k] = tProb16[indEdgesInPath[i]][j][k];
    }
    for(int j=0; j<19; j++){
      savedTProb18[i][j].resize(19);
      for(int k=0; k<19; k++)
	savedTProb18[i][j][k] = tProb18[indEdgesInPath[i]][j][k];
    }
  }
}



void Tree::setBackValues4()
{
  for(int i=0; i<numEdgesInPath; i++){
    *nodes[indNodesInPath[i]]=savedNodes[i];
    *edges[indEdgesInPath[i]]=savedEdges[i];
  }
  *nodes[indNodesInPath[numEdgesInPath]]=savedNodes[numEdgesInPath];

  for(int i=0; i<numMarkers; i++){
    for(int j=0; j<numNodesInPath2; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesInPath[j]][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesInPath[j]][k][m] = savedSubLogProb18[i][j][k][m];
      }
    for(int j=0; j<numNodesToRoot; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesToRoot[j]][k][m] = savedSubLogProb16[i][numNodesInPath2+j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesToRoot[j]][k][m] = savedSubLogProb18[i][numNodesInPath2+j][k][m];
      }

    for(int j=0; j<numEdgesInPath; j++)
      for(int k=0; k<2; k++)
	for(int m=0; m<2; m++)
	  tProb2[i][indEdgesInPath[j]][k][m] = savedTProb2[i][j][k][m];
  }
  for(int i=0; i<numEdgesInPath; i++){
    for(int j=0; j<17; j++)
      for(int k=0; k<17; k++)
	tProb16[indEdgesInPath[i]][j][k] = savedTProb16[i][j][k];
    for(int j=0; j<19; j++)
      for(int k=0; k<19; k++)
	tProb18[indEdgesInPath[i]][j][k] = savedTProb18[i][j][k];
  }
}

double Tree::calculateNewLogLikelihood4(const AFLPData &ad, const RunSettings& rs)
{
  logLikelihood = 0.0;
  for(int n=0; n<numEdgesInPath; n++) {
    int edgeNum = indEdgesInPath[n];
    Edge* edge0 = edges[edgeNum];
    edge0->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	setTProb16(edgeNum,i,j, edge0->getTProb16(i,j));
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	setTProb18(edgeNum,i,j, edge0->getTProb18(i,j));
    for(int k=0; k<numMarkers; k++){
      edge0->calculateTransitionMatricesC(ad.getMarkerLength(k));
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++)
	  setTProb2(k,edgeNum,i,j, edge0->getTProb2(i,j));	
    }
  }

  for(int i=0; i<numNodesInPath2; i++)
    calculateSubLogProb(nodes[indNodesInPathNew[i]],0, ad, rs);
  for(int i=0; i<numNodesToRoot; i++)
    calculateSubLogProb(nodes[indNodesToRoot[i]],0, ad, rs);

  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  if(numNodesToRoot>0 && indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2){
    cerr << " error: indNodesToRoot[numNodesToRoot-1]!=nodeNum1 && indNodesToRoot[numNodesToRoot-1]!=nodeNum2" << endl;
    exit(1);
  }
  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = getSubLogProb16(k, nodeNum1,0,0)+ getSubLogProb16(k,nodeNum2,0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = getSubLogProb16(k, nodeNum1,c1,m1)+ getSubLogProb16(k,nodeNum2,c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = getSubLogProb18(k, nodeNum1, c1,m1)+ getSubLogProb18(k,nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}

void Tree::setSavedValues5(int edgeChosen, int edgeAttached, int indCase)
{
  int numSavedEdges=3;
  if(indCase==1){
    // o
    //  \ /               o
    //   o              \/
    //   \/             o 
    //    o           \/
    //  e3\  /        o
    //atted\/        \/
    //      n2       o
    //       \   /  \/
    //      e1\ /e2 o
    //         n1  / e4  
    //       e0 \ / chosen
    //          n0                
    numSavedEdges=5;
    savedNodes.resize(3);
    savedEdges.resize(5);
    savedNodes[0] = *nodes[root];
    savedNodes[1] = (nodes[root]->getEdge(1)->getNumber()==edgeChosen ? *(nodes[root]->getNeighbor(2)) : *(nodes[root]->getNeighbor(1)));
    savedNodes[2] = *(edges[edgeAttached]->getNode(1));
    savedEdges[0] = *(savedNodes[1].getEdge(0));
    savedEdges[1] = *(savedNodes[1].getEdge(1));
    savedEdges[2] = *(savedNodes[1].getEdge(2));
    savedEdges[3] = *edges[edgeAttached];
    savedEdges[4] = *edges[edgeChosen];
    vector<int> nodesList(numNodes);
    findNodesToRoot(edges[edgeAttached]->getNode(0), nodesList, numNodesToRoot, 1, root);
    indNodesToRoot.resize(numNodesToRoot);
    for(int i=0; i<numNodesToRoot; i++)
      indNodesToRoot[i] = nodesList[numNodesToRoot-1-i];
    indNodesToRoot2.resize(numNodesToRoot);
    for(int i=1; i<numNodesToRoot-1; i++)
      indNodesToRoot2[i-1] = indNodesToRoot[i];
    indNodesToRoot2[numNodesToRoot-2] = indNodesToRoot[0];
    indNodesToRoot2[numNodesToRoot-1] = indNodesToRoot[numNodesToRoot-1];
  }
  else{   // indCase=2
    // o
    //  \ /               o
    //   o              \/
    //    \/   chosen\  o 
    //     o          \/e0
    //     \/         n0
    //      o        \/e1   
    //      \        n1
    //     e2\     \/
    //atted   \    o
    //         n2 /   
    //         \ /
    //          o                
    savedNodes.resize(3);
    savedEdges.resize(3);

    savedNodes[0] = *(edges[edgeChosen]->getNode(1));
    savedNodes[1] = *(savedNodes[0].getEdge(0)->getNode(1));
    /*    savedNodes[1] = *nodes[savedNodes[0].getNeighbor(0)->getNumber()];
    cout << edges[edgeChosen]->getNode(1)->getNeighbor(0)->getNumber() << endl;
    cout << (&(savedNodes[0]))->getNeighbor(0)->getNumber() << endl;        
    */
    savedNodes[2] = *(edges[edgeAttached]->getNode(1));
    savedEdges[0] = (savedNodes[0].getEdge(1)->getNumber()==edgeChosen ? *(savedNodes[0].getEdge(2)) : *(savedNodes[0].getEdge(1)));
    savedEdges[1] = *(savedNodes[0].getEdge(0));
    savedEdges[2] = *edges[edgeAttached];

    vector<int> nodesToRoot1(numNodes), nodesToRoot2(numNodes);
    int length1, length2;
    int include=0;
    findNodesToRoot(edges[edgeChosen]->getNode(0)->getSib(), nodesToRoot1, length1, 1, root);
    findNodesToRoot(edges[edgeAttached]->getNode(0), nodesToRoot2, length2, 1, root);
    indNodesToRoot.resize(numNodes);
    if(nodesToRoot1[length1-1]==nodesToRoot2[length2-1]){
      int lengthDiff = length2-length1;
      int i;
      for(i=length1-1; i>=0; i--){
	if(nodesToRoot1[i]==nodesToRoot2[i+lengthDiff])
	  indNodesToRoot[length1-1-i] = nodesToRoot1[i];
	else
	  break;
      }
      for(int j=i; j>=0; j--)
	indNodesToRoot[length1-1-j] = nodesToRoot1[j];
      if(lengthDiff+i<0)
	include=1;     // edgeAttached is one of the edges to root from e0's child not in the edgeChosen
      for(int j=lengthDiff+i; j>=0; j--)
	indNodesToRoot[length2+i-j] = nodesToRoot2[j];         // l1+(lD+i-j)=l2+i-j
      numNodesToRoot = length2 + i +1;
    }
    else{
      for(int i=0; i<length1; i++)
	indNodesToRoot[i] = nodesToRoot1[length1-1-i];
      for(int i=0; i<length2; i++)
	indNodesToRoot[length1+i] = nodesToRoot2[length2-1-i];
      numNodesToRoot = length1 + length2;
    }
    indNodesToRoot2.resize(numNodesToRoot);
    if(include==0){
      for(int i=0; i<length1-2; i++)
	indNodesToRoot2[i] = indNodesToRoot[i];
      for(int i=length1-1; i<numNodesToRoot-1; i++)
	indNodesToRoot2[i-1] = indNodesToRoot[i];
      indNodesToRoot2[numNodesToRoot-2] = indNodesToRoot[length1-2];
      indNodesToRoot2[numNodesToRoot-1] = indNodesToRoot[numNodesToRoot-1];
    }
    else{
      for(int i=0; i<length2-1; i++)
	indNodesToRoot2[i] = indNodesToRoot[i];
      indNodesToRoot2[length2-1] = indNodesToRoot[numNodesToRoot-2];
      for(int i=length2-1; i<numNodesToRoot-2; i++)
	indNodesToRoot2[i+1] = indNodesToRoot[i];
      indNodesToRoot2[numNodesToRoot-1] = indNodesToRoot[numNodesToRoot-1];
    }
  }

  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedSubLogProb16[i].resize(numNodesToRoot);
    savedSubLogProb18[i].resize(numNodesToRoot);
    for(int j=0; j<numNodesToRoot; j++){
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }
	
  savedTProb2.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedTProb2[i].resize(numSavedEdges);
    for(int j=0; j<numSavedEdges; j++){
      savedTProb2[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedTProb2[i][j][k].resize(2);
	for(int m=0; m<2; m++)
	  savedTProb2[i][j][k][m] = tProb2[i][savedEdges[j].getNumber()][k][m];
      }
    }
  }
  savedTProb16.resize(numSavedEdges);
  savedTProb18.resize(numSavedEdges);
  for(int i=0; i<numSavedEdges; i++){
    savedTProb16[i].resize(17);
    savedTProb18[i].resize(19);
    for(int j=0; j<17; j++){
      savedTProb16[i][j].resize(17);
      for(int k=0; k<17; k++)
	savedTProb16[i][j][k] = tProb16[savedEdges[i].getNumber()][j][k];
    }
    for(int j=0; j<19; j++){
      savedTProb18[i][j].resize(19);
      for(int k=0; k<19; k++)
	savedTProb18[i][j][k] = tProb18[savedEdges[i].getNumber()][j][k];
    }
  }
}


void Tree::setBackValues5(int indCase)
{
  int numSavedEdges = (indCase==1 ? 5:3);
  for(int i=0; i<3; i++)
    *nodes[savedNodes[i].getNumber()]=savedNodes[i];
  for(int i=0; i<numSavedEdges; i++)
    *edges[savedEdges[i].getNumber()]=savedEdges[i];

  for(int i=0; i<numMarkers; i++){
    for(int j=0; j<numNodesToRoot; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesToRoot[j]][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesToRoot[j]][k][m] = savedSubLogProb18[i][j][k][m];
      }
    for(int j=0; j<numSavedEdges; j++)
      for(int k=0; k<2; k++)
	for(int m=0; m<2; m++)
	  tProb2[i][savedEdges[j].getNumber()][k][m] = savedTProb2[i][j][k][m];
  }
  for(int i=0; i<numSavedEdges; i++){
    for(int j=0; j<17; j++)
      for(int k=0; k<17; k++)
	tProb16[savedEdges[i].getNumber()][j][k] = savedTProb16[i][j][k];
    for(int j=0; j<19; j++)
      for(int k=0; k<19; k++)
	tProb18[savedEdges[i].getNumber()][j][k] = savedTProb18[i][j][k];
  }
}

double Tree::calculateNewLogLikelihood5(const AFLPData &ad, const RunSettings& rs, int indCase)
{
  logLikelihood = 0.0;
  vector<int> indEdges;
  int size;
  if(indCase==1){
    size=2;
    indEdges.resize(2);
    indEdges[0]=savedEdges[0].getNumber();
    indEdges[1]=savedEdges[3].getNumber();
  }
  else{
    size=3;
    indEdges.resize(3);
    for(int i=0; i<3; i++)
      indEdges[i]=savedEdges[i].getNumber();
  }
  for(int n=0; n<size; n++) {
    int edgeNum = indEdges[n];
    Edge* edge0 = edges[edgeNum];
    edge0->calculateTransitionMatricesM();
    for(int i=0; i<17; i++)
      for(int j=0; j<17; j++)
	setTProb16(edgeNum,i,j, edge0->getTProb16(i,j));
    for(int i=0; i<19; i++)
      for(int j=0; j<19; j++)
	setTProb18(edgeNum,i,j, edge0->getTProb18(i,j));
    for(int k=0; k<numMarkers; k++){
      edge0->calculateTransitionMatricesC(ad.getMarkerLength(k));
      for(int i=0; i<2; i++)
	for(int j=0; j<2; j++)
	  setTProb2(k,edgeNum,i,j, edge0->getTProb2(i,j));	
    }
  }

  for(int i=numNodesToRoot-1; i>=0; i--)
    calculateSubLogProb(nodes[indNodesToRoot2[i]],0, ad, rs);

  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  if(numNodesToRoot>0 && indNodesToRoot2[0]!=nodeNum1 && indNodesToRoot2[0]!=nodeNum2){
    cerr << " error: indNodesToRoot2[0]!=nodeNum1 && indNodesToRoot2[0]!=nodeNum2" << endl;
    exit(1);
  }
  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = getSubLogProb16(k, nodeNum1,0,0)+ getSubLogProb16(k,nodeNum2,0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = getSubLogProb16(k, nodeNum1,c1,m1)+ getSubLogProb16(k,nodeNum2,c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = getSubLogProb18(k, nodeNum1, c1,m1)+ getSubLogProb18(k,nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}


void Tree::setSavedValues6(int edgeChosen1, int edgeChosen2, vector<int>& nodesToRoot1, int length1)
{
  savedEdges.resize(2);
  savedNodes.resize(2);
  savedEdges[0] = *edges[edgeChosen1];
  savedEdges[1] = *edges[edgeChosen2];
  savedNodes[0] = *(edges[edgeChosen1]->getNode(1));
  savedNodes[1] = *(edges[edgeChosen2]->getNode(1));

  vector<int> nodesToRoot2(numNodes);
  int length2;
  int include=0;
  findNodesToRoot(edges[edgeChosen2]->getNode(0), nodesToRoot2, length2, 0, root);
  indNodesToRoot.resize(numNodes);
  if(nodesToRoot1[length1-1]==nodesToRoot2[length2-1]){
    int lengthDiff = length2-length1;
    int i;
    for(i=length1-1; i>=0; i--){
      if(nodesToRoot1[i]==nodesToRoot2[i+lengthDiff])
	indNodesToRoot[length1-1-i] = nodesToRoot1[i];
      else
	break;
    }
    for(int j=i; j>=0; j--)
      indNodesToRoot[length1-1-j] = nodesToRoot1[j];
    if(lengthDiff+i<0)
      include=1;     
    for(int j=lengthDiff+i; j>=0; j--)
      indNodesToRoot[length2+i-j] = nodesToRoot2[j];         // l1+(lD+i-j)=l2+i-j
    numNodesToRoot = length2 + i +1;
  }
  else{
    for(int i=0; i<length1; i++)
      indNodesToRoot[i] = nodesToRoot1[length1-1-i];
    for(int i=0; i<length2; i++)
      indNodesToRoot[length1+i] = nodesToRoot2[length2-1-i];
    numNodesToRoot = length1 + length2;
  }

  savedSubLogProb16.resize(numMarkers);
  savedSubLogProb18.resize(numMarkers);
  for(int i=0; i<numMarkers; i++){
    savedSubLogProb16[i].resize(numNodesToRoot);
    savedSubLogProb18[i].resize(numNodesToRoot);
    for(int j=0; j<numNodesToRoot; j++){
      savedSubLogProb16[i][j].resize(2);
      savedSubLogProb18[i][j].resize(2);
      for(int k=0; k<2; k++){
	savedSubLogProb16[i][j][k].resize(17);
	savedSubLogProb18[i][j][k].resize(19);
	for(int m=0; m<17; m++)
	  savedSubLogProb16[i][j][k][m] = subLogProb16[i][indNodesToRoot[j]][k][m];
	for(int m=0; m<19; m++)
	  savedSubLogProb18[i][j][k][m] = subLogProb18[i][indNodesToRoot[j]][k][m];
      }
    }
  }	
}


void Tree::setBackValues6()
{
  for(int i=0; i<2; i++){
    *nodes[savedNodes[i].getNumber()]=savedNodes[i];
    *edges[savedEdges[i].getNumber()]=savedEdges[i];
  }
  for(int i=0; i<numMarkers; i++){
    for(int j=0; j<numNodesToRoot; j++)
      for(int k=0; k<2; k++){
	for(int m=0; m<17; m++)
	  subLogProb16[i][indNodesToRoot[j]][k][m] = savedSubLogProb16[i][j][k][m];
	for(int m=0; m<19; m++)
	  subLogProb18[i][indNodesToRoot[j]][k][m] = savedSubLogProb18[i][j][k][m];
      }
  }
}

double Tree::calculateNewLogLikelihood6(const AFLPData &ad, const RunSettings& rs)
{
  logLikelihood = 0.0;
  for(int i=numNodesToRoot-1; i>=0; i--)
    calculateSubLogProb(nodes[indNodesToRoot[i]],0, ad, rs);

  int nodeNum1 = nodes[root]->getNeighbor(1)->getNumber();
  int nodeNum2 = nodes[root]->getNeighbor(2)->getNumber();
  if(numNodesToRoot>0 && indNodesToRoot[0]!=nodeNum1 && indNodesToRoot[0]!=nodeNum2){
    cerr << " error: indNodesToRoot[0]!=nodeNum1 && indNodesToRoot[0]!=nodeNum2" << endl;
    exit(1);
  }

  double sum16, sum18, sum;
  double tuning;
  for(int k=0; k<numMarkers; k++){
    sum16=0.0;
    tuning = getSubLogProb16(k, nodeNum1,0,0)+ getSubLogProb16(k,nodeNum2,0,0);
    for(int c1=0;c1<2;c1++){
      for(int m1=0;m1<17;m1++) {
	double x = getSubLogProb16(k, nodeNum1,c1,m1)+ getSubLogProb16(k,nodeNum2,c1,m1);
	if(x > tuning)
	  tuning = x;
      }
      for(int m1=0;m1<19;m1++) {
	double x = getSubLogProb18(k, nodeNum1, c1,m1)+ getSubLogProb18(k,nodeNum2, c1,m1);
	if(x > tuning)
	  tuning = x;
      }
    }
    tuning *= -1;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<17;m++){
	sum += exp( log(piM16[m]) + getSubLogProb16(k, nodeNum1, c,m) + getSubLogProb16(k, nodeNum2, c,m) + tuning);      
      }
      sum16 += piC[k][c]*sum;
    }
    sum18=0.0;
    for(int c=0;c<2;c++) {
      sum = 0.0;
      for(int m=0;m<19;m++)
      sum += exp( log(piM18[m]) + getSubLogProb18(k, nodeNum1, c,m) + getSubLogProb18(k, nodeNum2, c,m) + tuning);   
      sum18 += piC[k][c]*sum;
    }
    logLikelihood += log((32*sum16+sum18)/33) -tuning;
  }
  return logLikelihood;
}


void Tree::updateTree(const RunSettings& rs,const AFLPData& ad, Rand& rand, int& move, int& accept, double& acceptProb)
{
  double newLogLikelihood;
  double logLikelihood1;
  move = (int)(rand.runif()*6);
  accept = 0;
  acceptProb=0;
  savedLogLikelihood = logLikelihood;

  if(move==0){    // multiply all edges by a variable
    //    double temp = -log(rand.runif());
    double alphaA = rs.getAlphaATuning();
    double temp = rand.rgamma(alphaA, alphaA);
    double sumEdges=0;
    for(int i=0; i<numEdges; i++){
      double tempLen = edges[i]->getLength();
      sumEdges += tempLen;
      edges[i]->setSavedLength();
      edges[i]->setLength(tempLen*temp);
    }
    setSavedProbs0();
    logLikelihood = calculateLogLikelihood(ad, rs, 0);
    //    acceptProb = exp( - (temp-1) * sumEdges / rs.getMu() + logLikelihood-savedLogLikelihood + temp-1/temp) * pow(temp, numEdges-2);    
    acceptProb = exp( - (temp-1) * sumEdges / rs.getMu() + logLikelihood-savedLogLikelihood + alphaA*(temp-1/temp)) / pow(temp, 2*alphaA-numEdges);    
    if(rand.runif() < acceptProb){
      accept = 1;
    }
    else{
      for(int i=0; i<numEdges; i++){
	edges[i]->setBackLength();
	logLikelihood = savedLogLikelihood;
	setBackProbs0();
      }
    }
  }

  else if(move==1){     // update a single length
    int EdgeChanged = (int)(rand.runif() * numEdges);
    //    double temp = -log(rand.runif());
    double alphaS = rs.getAlphaSTuning();
    double temp = rand.rgamma(alphaS, alphaS);
    double tempLen = edges[EdgeChanged]->getLength();
    edges[EdgeChanged]->setSavedLength();
    edges[EdgeChanged]->setSavedProb2();
    edges[EdgeChanged]->setLength(tempLen*temp);

    indNodesToRoot.resize(numEdges);          // exclude the root
    indEdgesToRoot.resize(numEdges);
    indNodesToRoot[0] = edges[EdgeChanged]->getNode(0)->getNumber();
    indEdgesToRoot[0] = EdgeChanged;
    numNodesToRoot = 0;
    Node* tempNode = edges[EdgeChanged]->getNode(1);
    while(tempNode->getNumber() != root){
      numNodesToRoot++;
      indNodesToRoot[numNodesToRoot] = tempNode->getNumber();
      indEdgesToRoot[numNodesToRoot] = tempNode->getEdge(0)->getNumber();
      tempNode = edges[indEdgesToRoot[numNodesToRoot]]->getNode(1);
    }
    numNodesToRoot++;
    setSavedProbs1(EdgeChanged);
    logLikelihood = calculateNewLogLikelihood1(ad,rs);

//    cout << move << "   loglike = "<< savedLogLikelihood << setw(8) << "  newlog = " << logLikelihood << endl;

//    acceptProb = exp( - (temp-1) * tempLen / rs.getMu() + logLikelihood-savedLogLikelihood + temp-1/temp) / temp;
    acceptProb = exp( - (temp-1) * tempLen / rs.getMu() + logLikelihood-savedLogLikelihood + alphaS*(temp-1/temp)) / pow(temp, 2*alphaS-1);
    if(rand.runif() < acceptProb){
      accept = 1;
    }
    else{
      edges[EdgeChanged]->setBackLength();
      edges[EdgeChanged]->setBackProb2();      
      logLikelihood = savedLogLikelihood;
      setBackProbs1(EdgeChanged);
    }
  }

  else if(move==2){       // reroot the tree
    int EdgeRooted = (int)(rand.runif() * (numEdges-1));
    int maxEdgeNum = nodes[root]->getEdge(1)->getNumber();
    int minEdgeNum = nodes[root]->getEdge(2)->getNumber();
    if(maxEdgeNum < minEdgeNum){
      int tempNum = minEdgeNum;
      minEdgeNum = maxEdgeNum;
      maxEdgeNum = tempNum;
    }
    if(EdgeRooted >= maxEdgeNum)
      EdgeRooted++;

    double tempLen4;
    tempLen4 = edges[maxEdgeNum]->getLength() + edges[minEdgeNum]->getLength();

    if(EdgeRooted != minEdgeNum){
      int indexEdge=EdgeRooted;             // index the edge-numbers directing to the old root
      double tempLen1, tempLen2, tempLen3;
      tempLen3 = edges[EdgeRooted]->getLength();
      tempLen1 = rand.runif()* tempLen3;
      tempLen2 = tempLen3 - tempLen1;

      acceptProb = tempLen3/tempLen4;

      if(rand.runif() < acceptProb){
	accept = 1;
	Node *nodesToRoot[numNodes];
	Edge *edgesToRoot[numEdges];
	int k=0;
	indexEdge=EdgeRooted;
	edgesToRoot[k] = edges[indexEdge];
	nodesToRoot[k] = edges[indexEdge]->getNode(1);
	while(nodesToRoot[k]->getNumber() != root){
	  indexEdge = nodesToRoot[k]->getEdge(0)->getNumber();
	  k++;
	  edgesToRoot[k] = edges[indexEdge];
	  nodesToRoot[k] = edgesToRoot[k]->getNode(1);
	}
	if(k==0){
	  cerr << "error: k=0;" << endl;
	  exit(1);
	}
	int j;
	for(j=1; j<3; j++){
	  if(nodesToRoot[k]->getEdge(j)==edgesToRoot[k]){
	    edgesToRoot[k+1] = nodesToRoot[k]->getEdge(3-j);
	    nodesToRoot[k+1] = edgesToRoot[k+1]->getNode(0);
	    break;
	  }
	}
 
	edgesToRoot[k+1]->setLength(tempLen4);
	edgesToRoot[k+1]->setNode(1,nodesToRoot[k-1]);
	Node *tempNode = nodes[root];
	Edge *tempEdge = edges[EdgeRooted];
	tempEdge->setNode(1,tempNode);
	tempEdge->setLength(tempLen1);
	edgesToRoot[0] = edgesToRoot[k];
	edgesToRoot[0]->setLength(tempLen2);
	edgesToRoot[0]->setNodes(nodesToRoot[0], tempNode);
	for(int i=1; i<k; i++){
	  edgesToRoot[i]->swapNodes();
	}
	for(int m=1; m<3; m++){
	  if(nodesToRoot[0]->getEdge(m)->getNumber()==EdgeRooted){
	    nodesToRoot[0]->setEdge(0, edgesToRoot[0]);
	    if(k==1)    //0==k-1
	      nodesToRoot[0]->setEdge(m, edgesToRoot[2]);
	    else
	      nodesToRoot[0]->setEdge(m, edgesToRoot[1]);
	    break;
	  }
	}
	for(int i=1; i<k; i++){
	  for(int m=1; m<3; m++){
	    if(nodesToRoot[i]->getEdge(m)->getNumber()==edgesToRoot[i]->getNumber()){
	      nodesToRoot[i]->setEdge(0,edgesToRoot[i]);
	      if(i==k-1)
		nodesToRoot[i]->setEdge(m,edgesToRoot[i+2]);
	      else
		nodesToRoot[i]->setEdge(m,edgesToRoot[i+1]);
	      break;
	    }
	  }
	}
	tempNode->setEdge(j,edgesToRoot[0]);
	tempNode->setEdge(3-j, tempEdge);
	calculateSubLogProb(nodesToRoot[k+1], 1, ad, rs);
	for(int i=k-1; i>0; i--)
	  calculateSubLogProb(nodesToRoot[i], 0, ad, rs);
	calculateSubLogProb(nodesToRoot[0], 1, ad, rs);
	calculateSubLogProb(tempEdge->getNode(0), 1, ad, rs);
	/*
	logLikelihood = calculateNewLogLikelihood2(ad,rs);
	if(abs(logLikelihood-savedLogLikelihood)>1.0e-10){
	  cerr << "loglikelihood != savedLoglikelihood" << endl;
	  exit(1);
	}
	*/
      }
    }
    else{           // always accepted
      double tempLen1 = rand.runif() * tempLen4;
      edges[minEdgeNum]->setLength(tempLen1);
      edges[maxEdgeNum]->setLength(tempLen4-tempLen1);
      accept = 1;
    }
  }

  else if(move==3){                   // move=3: local update
    int edgeChosen = (int)(rand.runif()*(numTaxa-3))+numTaxa;
    if(edgeChosen >= max(nodes[root]->getEdge(1)->getNumber(), nodes[root]->getEdge(2)->getNumber()))
      edgeChosen++;    

    Node *nodeU, *nodeV, *nodeA, *nodeB, *nodeC, *nodeD;
    Edge *e1, *e2, *e3, *e4, *e5, *e6;
    double distX, distY, distZ, distM, distMStar, distXStar, distYStar, distZStar;
    int tempInd1, tempInd2;
    int topology;
    int reject=0;

    nodeU = edges[edgeChosen]->getNode(0);
    tempInd1 = (int)(rand.runif()*2)+1;
    nodeA = nodeU->getNeighbor(tempInd1);
    nodeB = nodeU->getNeighbor(3-tempInd1);
    e1 = nodeA->getEdge(0);
    e2 = nodeB->getEdge(0);
    distX = e1->getLength();

    int j=0;
    for(j=1; j<3;j++)
      if(edgeChosen == nodes[root]->getEdge(j)->getNumber())
	break;

    if(j==1 || j==2){
      //    l   r  l   r   (a,b) = (l,r) or (r,l)
      //     \ /    \ /    (c,d) = (l,r) or (r,l)
      //      u      v        
      //       \\   /         
      //        \\ /
      //         p==root
      //      
      
      topology=0;
      nodeV = nodes[root]->getEdge(3-j)->getNode(0);
      if(nodeV->isLeaf())
	reject = 1;
      else{
	tempInd2 = (int)(rand.runif()*2)+1;
	nodeC = nodeV->getNeighbor(tempInd2);
	nodeD = nodeV->getNeighbor(3-tempInd2);
	e3 = nodeC->getEdge(0);
	e4 = nodeD->getEdge(0);
	e5 = nodeU->getEdge(0);
	e6 = nodeV->getEdge(0);
	distZ = distX + e5->getLength();
	distY = distZ + e6->getLength();
	distM = distY + e3->getLength();
      }
    }
    else{
      e5 = edges[edgeChosen];      //     nodeU->getEdge(0);
      nodeV = e5->getNode(1);
      distY = distX + e5->getLength();
      if(nodeV->getNeighbor(0)->getNumber()==root){

	//     
	//   l   r           (a,b) = (l,r) or (r,l)
	//    \ /            (c,d) = (s,o) or (o,s)
	//     u   s             
	//     \\ /              
	//       v   o         
	//        \ /
	//         root
	// 
	distZ = distY + nodeV->getEdge(0)->getLength();
	tempInd2 = (int)(rand.runif()*2);
	if(tempInd2==0){
	  topology=1;
	  nodeC = nodeU->getSib();
	  nodeD = nodeV->getSib();
	  e3 = nodeC->getEdge(0);
	  e4 = nodeD->getEdge(0);
	  e6 = nodeV->getEdge(0);
	  distM = distY + e3->getLength();
	}
	else{
	  topology=2;
	  nodeC = nodeV->getSib();
	  nodeD = nodeU->getSib();
	  e3 = nodeC->getEdge(0);
	  e4 = nodeD->getEdge(0);
	  e6 = nodeV->getEdge(0);
	  distM = distY + e3->getLength() + e6->getLength();
	}
      }
      else{                // nodeV->getNeighbor(0)->getNumber()!=root

	//   l   r           (a,b) = (l,r) or (r,l)
	//    \ /            (c,d) = (s,o) or (o,s)
	//     u   s             
	//     \\ /              
	//       v          
	//        \
	//         o!=root
	//
	distZ = 0;
	tempInd2 = (int)(rand.runif()*2);
	if(tempInd2==0){
	  topology=3;
	  nodeC = nodeU->getSib();
	  nodeD = nodeV->getNeighbor(0);
	  e3 = nodeC->getEdge(0);
	  e4 = nodeV->getEdge(0);
	}
	else{
	  topology=4;
	  nodeC = nodeV->getNeighbor(0);
	  nodeD = nodeU->getSib();
	  e3 = nodeV->getEdge(0);
	  e4 = nodeD->getEdge(0);
	}
	distM = distY + e3->getLength();
      }
    }

    if(topology<3){
      savedStates.NumSubEdges=6;
      savedStates.edges[0] = *e1;
      savedStates.edges[1] = *e2;
      savedStates.edges[2] = *e3;
      savedStates.edges[3] = *e4;
      savedStates.edges[4] = *e5;
      savedStates.edges[5] = *e6;
      savedStates.nodes[0] = *nodeA;
      savedStates.nodes[1] = *nodeB;
      savedStates.nodes[2] = *nodeC;
      savedStates.nodes[3] = *nodeD;
      savedStates.nodes[4] = *nodeU;
      savedStates.nodes[5] = *nodeV;
      savedStates.nodes[6] = *nodes[root];
    }
    else{
      savedStates.NumSubEdges=5;
      savedStates.edges[0] = *e1;
      savedStates.edges[1] = *e2;
      savedStates.edges[2] = *e3;
      savedStates.edges[3] = *e4;
      savedStates.edges[4] = *e5;
      savedStates.nodes[0] = *nodeA;
      savedStates.nodes[1] = *nodeB;
      savedStates.nodes[2] = *nodeC;
      savedStates.nodes[3] = *nodeD;
      savedStates.nodes[4] = *nodeU;
      savedStates.nodes[5] = *nodeV;
    }

    if(reject==0){       
      int nodeDetach;
      distMStar = distM * exp(rs.getLambdaLTuning()*(rand.runif()-.5));
      distZStar = distZ * distMStar / distM;
      if(rand.runif()<.5){
	nodeDetach = 0;
	distXStar = rand.runif()*distMStar;
	distYStar = distY * distMStar / distM;
      }
      else{
	nodeDetach = 1;
	distXStar = distX * distMStar / distM;
	distYStar = rand.runif() * distMStar;
      }/*
      cout << distXStar << setprecision(3) << setw(10) << distYStar << setprecision(3) << setw(10) << distZStar << endl;
      cout << nodeA->getNumber() << setw(5) << nodeC->getNumber() << endl;*/
      if(distXStar<distYStar){       
	if(topology==0){
	  numNodesToRoot = 4;
	  numEdgesToRoot = 4;
	  indNodesToRoot.resize(4);
	  indEdgesToRoot.resize(4);
	  if(distZStar<distXStar){
	    //  e1  e6   e5   e3  
	    // a---r---u----v----c
	    //       e2|    |e4
            //         b    d

	    e1->setNode(1, nodes[root]);
	    e1->setLength(distZStar);
	    e3->setLength(distMStar-distYStar);
	    e5->setNodes(nodeV, nodeU);
	    e5->setLength(distYStar-distXStar);
	    e6->setNodes(nodeU, nodes[root]);
	    e6->setLength(distXStar-distZStar);
	    nodeV->setEdge(0, e5);
	    nodeU->setEdge(0, e6);
	    for(int i=1; i<3; i++){
	      if(nodeU->getEdge(i)==e2){
		nodeU->setEdge(3-i, e5);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e1);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeC->getNumber();
	    indNodesToRoot[2] = nodeV->getNumber();
	    indNodesToRoot[3] = nodeU->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e3->getNumber();
	    indEdgesToRoot[2] = e5->getNumber();
	    indEdgesToRoot[3] = e6->getNumber();
	  }
	  else if(distZStar<distYStar){
	    //  e1     e5  e6   e3  
	    // a-----u---r---v----c
	    //     e2|       |e4
            //       b       d

	    e1->setLength(distXStar);
	    e3->setLength(distMStar-distYStar);
	    e5->setLength(distZStar-distXStar);
	    e6->setLength(distYStar-distZStar);
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeU->getNumber();
	    indNodesToRoot[2] = nodeC->getNumber();
	    indNodesToRoot[3] = nodeV->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	    indEdgesToRoot[3] = e6->getNumber();
	  }
	  else{
	    //  e1     e5   e6   e3  
	    // a-----u-----v---r---c
	    //     e2|     |e4
            //       b     d
		  
	    e1->setLength(distXStar);
	    e3->setNodes(nodeC, nodes[root]);
	    e3->setLength(distMStar-distZStar);
	    e5->setNodes(nodeU, nodeV);
	    e5->setLength(distYStar-distXStar);
	    e6->setNodes(nodeV, nodes[root]);
	    e6->setLength(distZStar-distYStar);
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e5);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e3);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeU->getNumber();
	    indNodesToRoot[2] = nodeV->getNumber();
	    indNodesToRoot[3] = nodeC->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e6->getNumber();
	    indEdgesToRoot[3] = e3->getNumber();
	  }
	}

	else if(topology==2){
	  numNodesToRoot = 4;
	  numEdgesToRoot = 4;
	  indNodesToRoot.resize(4);
	  indEdgesToRoot.resize(4);
	  if(distYStar<distZStar){
	    //  e1     e5   e6   e3  
	    // a-----u-----v---r---c
	    //     e2|     |e4
            //       b     d
		  
	    e1->setLength(distXStar);
	    e3->setLength(distMStar-distZStar);
	    e5->setLength(distYStar-distXStar);
	    e6->setLength(distZStar-distYStar);
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeU->getNumber();
	    indNodesToRoot[2] = nodeV->getNumber();
	    indNodesToRoot[3] = nodeC->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e6->getNumber();
	    indEdgesToRoot[3] = e3->getNumber();
	  }
	  else{
	    //   e1    e5   e6   e3  
	    // a-----u-----r---v---c
	    //     e2|         |e4
            //       b         d
		  
	    e1->setLength(distXStar);
	    e3->setNodes(nodeC, nodeV);
	    e3->setLength(distMStar-distYStar);
	    e5->setNodes(nodeU, nodes[root]);
	    e5->setLength(distZStar-distXStar);
	    e6->setNodes(nodeV, nodes[root]);
	    e6->setLength(distYStar-distZStar);
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e3);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e5);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeU->getNumber();
	    indNodesToRoot[2] = nodeC->getNumber();
	    indNodesToRoot[3] = nodeV->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	    indEdgesToRoot[3] = e6->getNumber();
	  }
	}

	else{   // topology != 0,2
	  e1->setLength(distXStar);
	  e3->setLength(distMStar-distYStar);
	  e5->setLength(distYStar-distXStar);
	  if(topology==1){
	    //   e1    e5   e3  
	    // a----u-----v----c
	    //    e2|   e6|
            //      b     r
	    //          e4|
	    //            d

	    numNodesToRoot = 4;
	    numEdgesToRoot = 3;
	    indNodesToRoot.resize(4);
	    indEdgesToRoot.resize(3);
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeU->getNumber();
	    indNodesToRoot[2] = nodeC->getNumber();
	    indNodesToRoot[3] = nodeV->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	  }
	  else if(topology==3){
	    //   e1    e5    e3  
	    // a----u----->v<----c
	    //    e2|    e6|
            //      b      V
	    //             d
	    numEdgesToRoot=3;
	    numEdgesToRoot = 3;
	    indEdgesToRoot.resize(3);
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	    indNodesToRoot.resize(numEdges);
	    indNodesToRoot[0] = nodeC->getNumber();
	    int tempNodeNum = nodeA->getNumber();
	    numNodesToRoot = 0;
	    while(tempNodeNum != root){
	      numNodesToRoot++;
	      indNodesToRoot[numNodesToRoot] = tempNodeNum;
	      tempNodeNum = nodes[tempNodeNum]->getNeighbor(0)->getNumber();
	    }
	    numNodesToRoot++;
	  }
	  else{
	    //   e1    e5    e3  
	    // a----u----->v---->c
	    //    e2|    e6|
            //      b      d
	    //             
	    numEdgesToRoot=3;
	    indEdgesToRoot.resize(3);
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	    indNodesToRoot.resize(numEdges);
	    int tempNodeNum = nodeA->getNumber();
	    numNodesToRoot = 0;
	    while(tempNodeNum != root){
	      indNodesToRoot[numNodesToRoot] = tempNodeNum;
	      tempNodeNum = nodes[tempNodeNum]->getNeighbor(0)->getNumber();
	      numNodesToRoot++;
	    }
	  }
	}
      }

      else{                          // distXStar>=distYStar, tree topology changes
	if(topology==0){
	  numNodesToRoot = 4;
	  numEdgesToRoot = 4;
	  indNodesToRoot.resize(4);
	  indEdgesToRoot.resize(4);
	  if(nodeDetach==0){      // nodeU is detached
	    //   e1    e6  e5   e3  
	    // a----r----v----u----c
	    //         e4|    |e2
            //           d    b
	    e1->setNodes(nodeA, nodes[root]);
	    e1->setLength(distZStar);
	    e3->setNodes(nodeC, nodeU);
	    e3->setLength(distMStar-distXStar);
	    e5->setNodes(nodeU, nodeV);
	    e5->setLength(distXStar-distYStar);
	    e6->setLength(distYStar-distZStar);
	    for(int i=1; i<3; i++){
	      if(nodeU->getEdge(i)==e2){
		nodeU->setEdge(3-i, e3);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e5);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e1);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeC->getNumber();
	    indNodesToRoot[2] = nodeU->getNumber();
	    indNodesToRoot[3] = nodeV->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e3->getNumber();
	    indEdgesToRoot[2] = e5->getNumber();
	    indEdgesToRoot[3] = e6->getNumber();
	  }
	  else{
	    //   e1   e5   e6   e3  
	    // a----v----u----r---c
	    //    e4|    |e2
            //      d    b
	    e1->setNodes(nodeA, nodeV);
	    e1->setLength(distYStar);
	    e3->setNodes(nodeC, nodes[root]);
	    e3->setLength(distMStar-distZStar);
	    e5->setNodes(nodeV, nodeU);
	    e5->setLength(distXStar-distYStar);
	    e6->setNodes(nodeU, nodes[root]);
	    e6->setLength(distZStar-distXStar);
	    nodeU->setEdge(0, e6);
	    for(int i=1; i<3; i++){
	      if(nodeU->getEdge(i)==e2){
		nodeU->setEdge(3-i, e5);
		break;
	      }
	    }
	    nodeV->setEdge(0, e5);
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e1);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e3);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeV->getNumber();
	    indNodesToRoot[2] = nodeU->getNumber();
	    indNodesToRoot[3] = nodeC->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e6->getNumber();
	    indEdgesToRoot[3] = e3->getNumber();
	  }
	}
	else if(topology==1){
	  //   e1    e5   e3  
	  // a-----v----u----c
	  //     e6|    |e2
	  //       r    b
	  //     e4|
	  //       d
	  e1->setNodes(nodeA, nodeV);
	  e1->setLength(distYStar);
	  e3->setNodes(nodeC, nodeU);
	  e3->setLength(distMStar-distXStar);
	  e5->setLength(distXStar-distYStar);
	  for(int i=1; i<3; i++){
	    if(nodeU->getEdge(i)==e2){
	      nodeU->setEdge(3-i, e3);
	      break;
	    }
	  }
	  for(int i=1; i<3; i++){
	    if(nodeV->getEdge(i)==e5){
	      nodeV->setEdge(3-i, e1);
	      break;
	    }
	  }
	  numNodesToRoot = 4;
	  numEdgesToRoot = 3;
	  indNodesToRoot.resize(4);
	  indEdgesToRoot.resize(3);
	  indNodesToRoot[0] = nodeA->getNumber();
	  indNodesToRoot[1] = nodeC->getNumber();
	  indNodesToRoot[2] = nodeU->getNumber();
	  indNodesToRoot[3] = nodeV->getNumber();
	  indEdgesToRoot[0] = e1->getNumber();
	  indEdgesToRoot[1] = e3->getNumber();
	  indEdgesToRoot[2] = e5->getNumber();
	}
	else if(topology==2){
	  numNodesToRoot = 4;
	  numEdgesToRoot = 4;
	  indNodesToRoot.resize(4);
	  indEdgesToRoot.resize(4);
	  if(distYStar<distXStar && distZStar<distXStar){
	    //   e1   e6   e5   e3  
	    // a----v----r----u---c
	    //    e4|         |e2
            //      d         b
	    e1->setNodes(nodeA, nodeV);
	    e1->setLength(distYStar);
	    e3->setNodes(nodeC, nodeU);
	    e3->setLength(distMStar-distXStar);
	    e5->setNodes(nodeU, nodes[root]);
	    e5->setLength(distXStar-distZStar);
	    e6->setNodes(nodeV, nodes[root]);
	    e6->setLength(distZStar-distYStar);
	    for(int i=1; i<3; i++){
	      if(nodeU->getEdge(i)==e2){
		nodeU->setEdge(3-i, e3);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e1);
		break;
	      }
	    }
	    for(int i=1; i<3; i++){
	      if(nodes[root]->getEdge(i)==e6){
		nodes[root]->setEdge(3-i, e5);
		break;
	      }
	    }	  
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeV->getNumber();
	    indNodesToRoot[2] = nodeC->getNumber();
	    indNodesToRoot[3] = nodeU->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e6->getNumber();
	    indEdgesToRoot[2] = e3->getNumber();
	    indEdgesToRoot[3] = e5->getNumber();
	  }
	  else if(distYStar<distXStar && distZStar>distXStar){
	    //   e1   e5   e6   e3  
	    // a----v----u----r---c
	    //    e4|    |e2
            //      d    b
	    e1->setNodes(nodeA, nodeV);
	    e1->setLength(distYStar);
	    e3->setLength(distMStar-distZStar);
	    e5->setNodes(nodeV, nodeU);
	    e5->setLength(distXStar-distYStar);
	    e6->setNodes(nodeU, nodes[root]);
	    e6->setLength(distZStar-distXStar);
	    nodeU->setEdge(0, e6);
	    for(int i=1; i<3; i++){
	      if(nodeU->getEdge(i)==e2){
		nodeU->setEdge(3-i, e5);
		break;
	      }
	    }
	    nodeV->setEdge(0, e5);
	    for(int i=1; i<3; i++){
	      if(nodeV->getEdge(i)==e4){
		nodeV->setEdge(3-i, e1);
		break;
	      }
	    }
	    indNodesToRoot[0] = nodeA->getNumber();
	    indNodesToRoot[1] = nodeV->getNumber();
	    indNodesToRoot[2] = nodeU->getNumber();
	    indNodesToRoot[3] = nodeC->getNumber();
	    indEdgesToRoot[0] = e1->getNumber();
	    indEdgesToRoot[1] = e5->getNumber();
	    indEdgesToRoot[2] = e6->getNumber();
	    indEdgesToRoot[3] = e3->getNumber();
	  }
	  else{
	    cout << "error: ";
	    exit(1);
	  }
	}
	else if(topology==3){
	  //   e1     e5   e3  
	  // a---->v<-----u---c
	  //     e4|      |e2
	  //       V      b
	  //       d
	  e1->setNodes(nodeA, nodeV);
	  e1->setLength(distYStar);
	  e3->setNodes(nodeC, nodeU);
	  e3->setLength(distMStar-distXStar);
	  e5->setLength(distXStar-distYStar);
	  for(int i=1; i<3; i++){
	    if(nodeU->getEdge(i)==e2){
	      nodeU->setEdge(3-i, e3);
	      break;
	    }
	  }
	  for(int i=1; i<3; i++){
	    if(nodeV->getEdge(i)==e5){
	      nodeV->setEdge(3-i, e1);
	      break;
	    }
	  }	  
	  numEdgesToRoot = 3;
	  indEdgesToRoot.resize(3);
	  indEdgesToRoot[0] = e1->getNumber();
	  indEdgesToRoot[1] = e3->getNumber();
	  indEdgesToRoot[2] = e5->getNumber();
	  indNodesToRoot.resize(numEdges);
	  indNodesToRoot[0] = nodeA->getNumber();
	  int tempNodeNum = nodeC->getNumber();
	  numNodesToRoot = 0;
	  while(tempNodeNum != root){
	    numNodesToRoot++;
	    indNodesToRoot[numNodesToRoot] = tempNodeNum;
	    tempNodeNum = nodes[tempNodeNum]->getNeighbor(0)->getNumber();
	  }
	  numNodesToRoot++;
	}
	else{            // topology=4
	  //   e1    e5     e3  
	  // a---->v----->u---->c
	  //     e4|      |e2
	  //       d      b
	  e1->setNodes(nodeA, nodeV);
	  e1->setLength(distYStar);
	  e3->setNodes(nodeU, nodeC);
	  e3->setLength(distMStar-distXStar);
	  e5->setNodes(nodeV, nodeU);
	  e5->setLength(distXStar-distYStar);
	  nodeU->setEdge(0, e3);
	  for(int i=1; i<3; i++){
	    if(nodeU->getEdge(i)==e2){
	      nodeU->setEdge(3-i, e5);
	      break;
	    }
	  }
	  nodeV->setEdge(0, e5);
	  for(int i=1; i<3; i++){
	    if(nodeV->getEdge(i)==e4){
	      nodeV->setEdge(3-i, e1);
	      break;
	    }
	  }
	  numEdgesToRoot=3;
	  indEdgesToRoot.resize(3);
	  indEdgesToRoot[0] = e1->getNumber();
	  indEdgesToRoot[1] = e5->getNumber();
	  indEdgesToRoot[2] = e3->getNumber();
	  indNodesToRoot.resize(numEdges);
	  int tempNodeNum = nodeA->getNumber();	  
	  numNodesToRoot = 0;
	  while(tempNodeNum != root){
	    indNodesToRoot[numNodesToRoot] = tempNodeNum;
	    tempNodeNum = nodes[tempNodeNum]->getNeighbor(0)->getNumber();
	    numNodesToRoot++;
	  }
	}
      }          

      setSavedProbs3();
      logLikelihood = calculateNewLogLikelihood3(ad,rs);
      /*
      logLikelihoodX = calculateLogLikelihoodX(ad,rs);
      if(fabs(logLikelihood -logLikelihoodX) > 1.0e-10) {
	cerr << "Check Error: Likelihoods differ after proposal 3" << endl;
	cerr << "logLikelihood  = " << logLikelihood << endl;
	cerr << "logLikelihoodX = " << logLikelihoodX << endl;
	for(int k=0; k<numMarkers; k++)
	  for(int i=0; i<numEdges; i++){
	    for(int j=0; j<2; j++)
	      for(int m=0; m<2; m++)
		if(fabs(tProb2[k][i][j][m] - tProb2X[k][i][j][m]) > 1.0e-10){
		  cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		  cerr << " tProb2: " << tProb2[k][i][j][m] << "  tProb2X: " << tProb2X[k][i][j][m] << endl;
		}
	  }
	for(int i=0; i<numEdges; i++){	  
	  for(int j=0; j<17; j++)
	    for(int m=0; m<17; m++)
	      if(fabs(tProb16[i][j][m] - tProb16X[i][j][m]) > 1.0e-10){
		cerr << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		cerr << " tProb16: " << tProb16[i][j][m] << "  tProb16X: " << tProb16X[i][j][m] << endl;
	      }
	  for(int j=0; j<19; j++)
	    for(int m=0; m<19; m++)
	      if(fabs(tProb18[i][j][m] - tProb18X[i][j][m]) > 1.0e-10){
		cerr << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		cerr << " tProb18: " << tProb18[i][j][m] << "  tProb18X: " << tProb18X[i][j][m] << endl;
	      }
	}
	for(int k=0; k<numMarkers; k++)
	  for(int i=0; i<numEdges; i++)
	    for(int j=0; j<2; j++){
	      for(int m=0; m<17; m++)
		if(fabs(subLogProb16[k][i][j][m] - subLogProb16X[k][i][j][m]) > 1.0e-10){
		  cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		  cerr << " subLogProb16: " << subLogProb16[k][i][j][m] << "  subLogProb16X: " << subLogProb16X[k][i][j][m] << endl;
		}
	      for(int m=0; m<19; m++)
		if(fabs(subLogProb18[k][i][j][m] - subLogProb18X[k][i][j][m]) > 1.0e-10){
		  cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		  cerr << " subLogProb18: " << subLogProb18[k][i][j][m] << "  subLogProb18X: " << subLogProb18X[k][i][j][m] << endl;
		}
	    }    
	exit(1);
      }
      */
      if(topology==0 || topology==2)
	acceptProb = exp(-(distMStar-distM)/rs.getMu() + logLikelihood - savedLogLikelihood) * pow(distMStar/distM, 4);
      else
	acceptProb = exp(-(distMStar-distM)/rs.getMu() + logLikelihood - savedLogLikelihood) * pow(distMStar/distM, 3);

      if(rand.runif()<acceptProb){
	accept=1;
      }
      else{
	for(int i=0; i<savedStates.NumSubEdges; i++){
	  *edges[(savedStates.edges[i]).getNumber()] = (savedStates.edges[i]);
	  *nodes[(savedStates.nodes[i]).getNumber()] = (savedStates.nodes[i]);
	}
	*nodes[(savedStates.nodes[savedStates.NumSubEdges]).getNumber()] 
                             = (savedStates.nodes[savedStates.NumSubEdges]);
	setBackProbs3();
	logLikelihood = savedLogLikelihood;
      }
    }
  }

  else if(move==4){
    // move = 4: generalized local update
    // pick two leaves, stretch the length between them, random remove one edge and stick it in a random place along the path

    int IndNode1 = (int)(numTaxa * rand.runif());
    int IndNode2  = (int)((numTaxa-1) * rand.runif());
    if(IndNode2>=IndNode1) IndNode2++;
    //    cout << "IndNode1= " << IndNode1 << "  IndNode2= " << IndNode2 << endl;
    searchPath(nodes[IndNode1], nodes[IndNode2]);
    if(indNodesInPath[numNodesInPath-1]!=root){      // find indNodesToRoot
      numNodesInPath2=numNodesInPath;
      indNodesToRoot.resize(numNodes);
      indEdgesToRoot.resize(numEdges);
      int n=nodes[indNodesInPath[numNodesInPath-1]]->getNeighbor(0)->getNumber();
      numNodesToRoot=0;
      while(n!=root){
	indNodesToRoot[numNodesToRoot]=n;
	n=nodes[n]->getNeighbor(0)->getNumber();
	numNodesToRoot++;
      }
    }
    else{
      numNodesInPath2=numNodesInPath-1;
      numNodesToRoot=0;
    }   

    vector<int> indNodesInPath2(numNodesInPath);          // in the order starting from nodeS to nodeE
    vector<int> indEdgesInPath2(numEdgesInPath);
    int intAnc;
    indNodesInPath2[0]=IndNode1;
    indEdgesInPath2[0]=indEdgesInPath[0];
    for(int i=1; i<numEdgesInPath; i++){
      if(indNodesInPath[i]<numTaxa){
	int k=0;
	for(int j=numEdgesInPath-1; j>=i; j--){
	  indNodesInPath2[i+k]=indNodesInPath[j+1];
	  indEdgesInPath2[i+k]=indEdgesInPath[j];
	  k++;
	}
	indNodesInPath2[numEdgesInPath]=IndNode2;
	intAnc=i;
	break;
      }
      else{
	indNodesInPath2[i]=indNodesInPath[i];
	indEdgesInPath2[i]=indEdgesInPath[i];
      }
    }
    /*
    cout << "the path between them is:" << endl;
    for(int i=0; i<numNodesInPath; i++)
      cout << indNodesInPath[i] << setw(6) << indNodesInPath2[i] << endl;
    cout << "the numNodesInPath = " << numNodesInPath << endl;
    */
    setSavedValues4();

    int tempInt = (int)((numNodesInPath-2) * rand.runif() + 1);
    //    cout << "indNodeChosen= " << indNodesInPath2[tempInt] << endl;
    double dist[numEdgesInPath], distStar[numEdgesInPath];
    double distRatio=exp(rs.getLambdaGTuning()*(rand.runif()-.5));
    dist[0]=edges[indEdgesInPath2[0]]->getLength();
    distStar[0]=dist[0]*distRatio;
    edges[indEdgesInPath2[0]]->setLength(distStar[0]);
    for(int i=1; i<numEdgesInPath; i++){
      dist[i]=dist[i-1]+edges[indEdgesInPath2[i]]->getLength();
      distStar[i] = dist[i]*distRatio;
      edges[indEdgesInPath2[i]]->setLength(distStar[i]-distStar[i-1]);
    }
    double tempDist=distStar[numEdgesInPath-1]*rand.runif();
    /*
    cout << "numNodesInPath= " << numNodesInPath << endl;
    for(int i=0; i<numEdgesInPath; i++)
      cout << dist[i] << setw(20) << distStar[i] << endl;
    cout << "tempDist= " << tempDist << endl;
    cout << "tempInt= " << tempInt << endl;
    */

    vector<int> indNodesInPathNew2(numNodesInPath);       // in the order starting from nodeS to nodeE after update
    indNodesInPathNew.resize(numNodesInPath);
    for(int i=0; i<numNodesInPath; i++)
      indNodesInPathNew2[i]=indNodesInPath2[i];

    int indLowbd=tempInt-2, indUpbd=tempInt+1;
    if(tempDist<(tempInt>1? distStar[indLowbd]:0)){
      if(tempDist<distStar[0])
	indLowbd=0;
      else
	for(int i=tempInt-3; i>=0; i--)
	  if(tempDist>distStar[i]){
	    indLowbd=i+1;
	    break;
	  }

      for(int i=tempInt-1; i>indLowbd; i--)
	indNodesInPathNew2[i+1]=indNodesInPathNew2[i];
      indNodesInPathNew2[indLowbd+1]=indNodesInPath2[tempInt];

      if(indLowbd==0)
	edges[indEdgesInPath2[indLowbd]]->setLength(tempDist);
      else
	edges[indEdgesInPath2[indLowbd]]->setLength(tempDist-distStar[indLowbd-1]);

      if(tempInt<intAnc){
	// s--n1---n2---n3--n4---anc---
	// s--n3------n1---n2------n4--anc---
	//     tempInt      tempInt-1
	//	cout << "case 1: tempDist<distStar, tempInt<intAnc" << endl;
	for(int i=tempInt-1; i>indLowbd; i--)
	  indNodesInPathNew[i+1]=indNodesInPathNew[i];
	indNodesInPathNew[indLowbd+1]=indNodesInPathNew[tempInt];
	  
	edges[indEdgesInPath2[indLowbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	edges[indEdgesInPath2[tempInt]]->setNode(1, nodes[indNodesInPath2[indLowbd+1]]);
	edges[indEdgesInPath2[tempInt-1]]->setNode(1, nodes[indNodesInPath2[tempInt+1]]);
	edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	edges[indEdgesInPath2[tempInt]]->setLength(distStar[indLowbd]-tempDist);	
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	    nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indLowbd]]);
	    break;
	  }
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[indLowbd+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indLowbd]]->getNumber()){
	    nodes[indNodesInPath2[indLowbd+1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	    break;
	  }
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[tempInt+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	    nodes[indNodesInPath2[tempInt+1]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	    break;
	  }
      }
      else if(tempInt==intAnc){
	// s--n1---n2---n3--n4---anc---
	// s--n1--anc-----------n2--n3--n4---
	//           tempInt-1             tempInt 	

	if(tempInt>indLowbd+2){
	  // cout << "case 21: tempDist<distStar, tempInt=intAnc" << endl;
	  edges[indEdgesInPath2[indLowbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt-1]]->setNode(0, nodes[indNodesInPath2[indLowbd+1]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1, nodes[indNodesInPath2[tempInt-1]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[indLowbd]-tempDist);	
	  for(int i=indLowbd+1; i<tempInt-1; i++)
	    edges[indEdgesInPath2[i]]->swapNodes();
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(3-i,edges[indEdgesInPath2[indLowbd]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indLowbd+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indLowbd]]->getNumber()){
	      nodes[indNodesInPath2[indLowbd+1]]->setEdge(0,edges[indEdgesInPath2[tempInt-1]]);
	      nodes[indNodesInPath2[indLowbd+1]]->setEdge(i,edges[indEdgesInPath2[indLowbd+1]]);
	      break;
	    }
	  for(int j=indLowbd+2; j<tempInt-1; j++)
	    for(int i=1; i<3; i++)
	      if(nodes[indNodesInPath2[j]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[j-1]]->getNumber()){
		nodes[indNodesInPath2[j]]->setEdge(0,edges[indEdgesInPath2[j-1]]);
		nodes[indNodesInPath2[j]]->setEdge(i,edges[indEdgesInPath2[j]]);
		break;
	      }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt-1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-2]]->getNumber()){
	      nodes[indNodesInPath2[tempInt-1]]->setEdge(0,edges[indEdgesInPath2[tempInt-2]]);
	      nodes[indNodesInPath2[tempInt-1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	}
	else{                // tempInt-1=indLowbd+1
	  // s--n1---------n2---anc---
	  // s--n1--anc-----------n2--
	  //           tempInt-1     tempInt 	
	  // cout << "case 22: tempDist<distStar, tempInt=intAnc" << endl;
	  edges[indEdgesInPath2[indLowbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1, nodes[indNodesInPath2[tempInt-1]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[indLowbd]-tempDist);	
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(3-i,edges[indEdgesInPath2[indLowbd]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indLowbd+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indLowbd]]->getNumber()){
	      nodes[indNodesInPath2[indLowbd+1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	}
      }
      else{             // tempInt>intAnc
	if(indLowbd<intAnc){
	  // s--n1---n2---n3---anc---n4---n5---
	  // s--n1--n2--n4--------n3---anc--------n5---
	  //              tempInt-1        tempInt                 
	  // s--n1--n2--n3--n4--------anc--------n5
	  //                  tempInt-1      tempInt
	  // cout << "case 31: tempDist<distStar, tempInt>intAnc" << endl;
	  edges[indEdgesInPath2[indLowbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1, nodes[indNodesInPath2[indLowbd+1]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1, nodes[indNodesInPath2[tempInt-1]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[indLowbd]-tempDist);	
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indLowbd]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indLowbd+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indLowbd]]->getNumber()){
	      nodes[indNodesInPath2[indLowbd+1]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt-1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt-1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	}
	else{
	  // s--n1---anc----n2---n3--n4---n5---
	  // s--n1---anc--------n5-- n2---n3--n4----
	  //            tempInt-1               tempInt
	  // cout << "case 32: tempDist<distStar, tempInt>intAnc" << endl;
	  edges[indEdgesInPath2[indLowbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[indLowbd]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1,nodes[indNodesInPath2[tempInt-1]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  if(indLowbd>0)
	    edges[indEdgesInPath2[tempInt-1]]->setLength(tempDist-distStar[indLowbd-1]);
	  else
	    edges[indEdgesInPath2[tempInt-1]]->setLength(tempDist);
	  edges[indEdgesInPath2[indLowbd]]->setLength(distStar[indLowbd]-tempDist);
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indLowbd]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indLowbd]]->getNumber()){
	      nodes[indNodesInPath2[indLowbd]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indLowbd]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt-1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt-1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	}
      }
    }

    else if(tempDist>distStar[tempInt]){     
      for(int i=tempInt+1; i<numNodesInPath; i++)
	if(tempDist<distStar[i]){
	  indUpbd=i;
	  break;
	}
      for(int i=tempInt+1; i<=indUpbd; i++)
	indNodesInPathNew2[i-1]=indNodesInPathNew2[i];
      indNodesInPathNew2[indUpbd]=indNodesInPath2[tempInt];

      edges[indEdgesInPath2[indUpbd]]->setLength(distStar[indUpbd]-tempDist);

      if(tempInt>intAnc){
	// s--n1---anc----n2---n3--n4---n5---
	// s--n1---anc---------n3--n4--n5---------n2--
	//            tempInt             tempInt-1   
	// cout << "case 4: tempDist>distStar, tempInt>intAnc" << endl;
	edges[indEdgesInPath2[indUpbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[indUpbd]]);
	edges[indEdgesInPath2[tempInt]]->setNode(1,nodes[indNodesInPath2[tempInt-1]]);
	edges[indEdgesInPath2[tempInt-1]]->setLength(tempDist-distStar[indUpbd-1]);
	edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[tempInt-1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	    nodes[indNodesInPath2[tempInt-1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	    break;
	  }
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[indUpbd]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indUpbd]]->getNumber()){
	    nodes[indNodesInPath2[indUpbd]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	    break;
	  }
	for(int i=1; i<3; i++)
	  if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	    nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indUpbd]]);
	    break;
	  }
      }
      else if(tempInt==intAnc){
	// s--n1---anc---n2---n3--n4---n5---
	// s--n1-------n2---n3--n4------anc--n5---
	//     tempInt-1         tempInt 
	if(indUpbd>tempInt+1){
	  // cout << "case 51: tempDist>distStar, tempInt=intAnc" << endl;
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[tempInt+1]]);
	  for(int i=tempInt+1; i<indUpbd; i++)
	    edges[indEdgesInPath2[i]]->swapNodes();
	  edges[indEdgesInPath2[tempInt]]->setNode(0,nodes[indNodesInPath2[indUpbd]]);
	  edges[indEdgesInPath2[indUpbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(tempDist-distStar[indUpbd-1]);
	  if(tempInt>1)
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  else
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]);
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt+1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt+1]]->setEdge(0,edges[indEdgesInPath2[tempInt+1]]);
	      nodes[indNodesInPath2[tempInt+1]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	    }
	  for(int j=tempInt+2; j<indUpbd; j++)
	    for(int i=1; i<3; i++)
	      if(nodes[indNodesInPath2[j]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[j]]->getNumber()){
		nodes[indNodesInPath2[j]]->setEdge(0,edges[indEdgesInPath2[j]]);
		nodes[indNodesInPath2[j]]->setEdge(i,edges[indEdgesInPath2[j-1]]);
		break;
	      }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indUpbd]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indUpbd]]->getNumber()){
	      nodes[indNodesInPath2[indUpbd]]->setEdge(i,edges[indEdgesInPath2[indUpbd-1]]);
	      nodes[indNodesInPath2[indUpbd]]->setEdge(0,edges[indEdgesInPath2[tempInt]]);

	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(3-i,edges[indEdgesInPath2[indUpbd]]);
	      break;
	  }
	}
	else{
	  // s--n1---anc---n2---
	  // s--n1-------n2-----anc--
	  //     tempInt-1  tempInt 
	  // cout << "case 52: tempDist>distStar, tempInt=intAnc" << endl;
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[tempInt+1]]);
	  edges[indEdgesInPath2[indUpbd]]->setNode(1, nodes[indNodesInPath2[tempInt]]);
	  if(tempInt>1)
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  else
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]);
	  edges[indEdgesInPath2[tempInt]]->setLength(tempDist-distStar[tempInt]);
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(3-i,edges[indEdgesInPath2[indUpbd]]);
	      break;
	  }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indUpbd]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indUpbd]]->getNumber()){
	      nodes[indNodesInPath2[indUpbd]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	  }
	}
      }
      else{             // tempInt<intAnc          
	if(indUpbd<intAnc){
	  // s--n1---n2---n3---anc---n4-
	  // s-------n2--n3--n1--------anc--n4-
	  // tempInt-1         tempInt
	  // cout << "case 61: tempDist>distStar, tempInt<intAnc" << endl;
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[tempInt+1]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1,nodes[indNodesInPath2[indUpbd+1]]);
	  edges[indEdgesInPath2[indUpbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[indUpbd]]->setLength(tempDist-distStar[indUpbd-1]);
	  edges[indEdgesInPath2[tempInt]]->setLength(distStar[indUpbd]-tempDist);
	  if(tempInt>1)
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  else
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]);
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indUpbd]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt+1]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indUpbd+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indUpbd]]->getNumber()){
	      nodes[indNodesInPath2[indUpbd+1]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	}
	else{                  // indUpbd>=intAnc
	  // s--n1--n2----anc--n3---n4-
	  // s-------n2---anc--n3------n1-------n4-
	  // tempInt-1           tempInt  indUpbd
	  // s--n2---anc--n1--n3--n4-
	  //
	  // cout << "case 62: tempDist>distStar, tempInt<intAnc" << endl;
	  edges[indEdgesInPath2[tempInt-1]]->setNode(1,nodes[indNodesInPath2[tempInt+1]]);
	  edges[indEdgesInPath2[tempInt]]->setNode(1,nodes[indNodesInPath2[indUpbd]]);
	  edges[indEdgesInPath2[indUpbd]]->setNode(1,nodes[indNodesInPath2[tempInt]]);
	  edges[indEdgesInPath2[tempInt]]->setLength(tempDist-distStar[indUpbd-1]);
	  if(tempInt>1)
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]-distStar[tempInt-2]);
	  else
	    edges[indEdgesInPath2[tempInt-1]]->setLength(distStar[tempInt]);
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt+1]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt]]->getNumber()){
	      nodes[indNodesInPath2[tempInt+1]]->setEdge(i,edges[indEdgesInPath2[tempInt-1]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[indUpbd]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[indUpbd]]->getNumber()){
	      nodes[indNodesInPath2[indUpbd]]->setEdge(i,edges[indEdgesInPath2[tempInt]]);
	      break;
	    }
	  for(int i=1; i<3; i++)
	    if(nodes[indNodesInPath2[tempInt]]->getEdge(i)->getNumber()==edges[indEdgesInPath2[tempInt-1]]->getNumber()){
	      nodes[indNodesInPath2[tempInt]]->setEdge(i,edges[indEdgesInPath2[indUpbd]]);
	      break;
	    }
	}
      }
    }

    else{      // distStar[tempInt-2]<tempDist<distStar[tempInt]:  topology unchanged
      // cout << "case 7: a<tempDist<b" << endl;
      if(tempInt>1)
	edges[indEdgesInPath2[tempInt-1]]->setLength(tempDist-distStar[tempInt-2]);
      else
	edges[indEdgesInPath2[tempInt-1]]->setLength(tempDist);
      edges[indEdgesInPath2[tempInt]]->setLength(distStar[tempInt]-tempDist);
    }

    int temp=indNodesInPath[numEdgesInPath];
    for(int i=0; i<numNodesInPath; i++){
      if(indNodesInPathNew2[i]==temp){
	for(int j=i; j<=numEdgesInPath; j++)
	  indNodesInPathNew[j]=indNodesInPathNew2[numEdgesInPath-(j-i)];
	break;
      }
      indNodesInPathNew[i]=indNodesInPathNew2[i];
    }

    logLikelihood = calculateNewLogLikelihood4(ad,rs);
    /*
    logLikelihoodX = calculateLogLikelihoodX(ad,rs);
    if(fabs(logLikelihood -logLikelihoodX) > 1.0e-10) {
      cerr << "Check Error: Likelihoods differ after proposal 4" << endl;
      cerr << "logLikelihood  = " << logLikelihood << endl;
      cerr << "logLikelihoodX = " << logLikelihoodX << endl;
      for(int k=0; k<numMarkers; k++)
	for(int i=0; i<numEdges; i++){
	  for(int j=0; j<2; j++)
	    for(int m=0; m<2; m++)
	      if(fabs(tProb2[k][i][j][m] - tProb2X[k][i][j][m]) > 1.0e-10){
		cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		cerr << " tProb2: " << tProb2[k][i][j][m] << "  tProb2X: " << tProb2X[k][i][j][m] << endl;
	      }
	}
      for(int i=0; i<numEdges; i++){	  
	for(int j=0; j<17; j++)
	  for(int m=0; m<17; m++)
	    if(fabs(tProb16[i][j][m] - tProb16X[i][j][m]) > 1.0e-10){
	      cerr << " edge: " << i << " and row: " << j << "  col: " << m << endl;
	      cerr << " tProb16: " << tProb16[i][j][m] << "  tProb16X: " << tProb16X[i][j][m] << endl;
	    }
	for(int j=0; j<19; j++)
	  for(int m=0; m<19; m++)
	    if(fabs(tProb18[i][j][m] - tProb18X[i][j][m]) > 1.0e-10){
	      cerr << " edge: " << i << " and row: " << j << "  col: " << m << endl;
	      cerr << " tProb18: " << tProb18[i][j][m] << "  tProb18X: " << tProb18X[i][j][m] << endl;
	    }
      }
      for(int k=0; k<numMarkers; k++)
	for(int i=0; i<numEdges; i++)
	  for(int j=0; j<2; j++){
	    for(int m=0; m<17; m++)
	      if(fabs(subLogProb16[k][i][j][m] - subLogProb16X[k][i][j][m]) > 1.0e-10){
		cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		cerr << " subLogProb16: " << subLogProb16[k][i][j][m] << "  subLogProb16X: " << subLogProb16X[k][i][j][m] << endl;
	      }
	    for(int m=0; m<19; m++)
	      if(fabs(subLogProb18[k][i][j][m] - subLogProb18X[k][i][j][m]) > 1.0e-10){
		cerr << " marker: " << k << " edge: " << i << " and row: " << j << "  col: " << m << endl;
		cerr << " subLogProb18: " << subLogProb18[k][i][j][m] << "  subLogProb18X: " << subLogProb18X[k][i][j][m] << endl;
	      }
	  }    
      exit(1);
    }
    */
    acceptProb = exp(-(distStar[numEdgesInPath-1]-dist[numEdgesInPath-1])/rs.getMu() + logLikelihood - savedLogLikelihood) * pow(distRatio, numEdgesInPath);

    if(rand.runif()<acceptProb){
      accept=1;
    }
    else{
      setBackValues4();
      logLikelihood = savedLogLikelihood;
    }
  }

  else if(move==5){                  // move = 5: SPR--subtree pruning and regrafting
    //    int edgeChosen = (int)(rand.runif()*numEdges);
    int edgeChosen = (int)(rand.runif()*(numEdges-2));
    int minEdge = nodes[root]->getEdge(1)->getNumber();
    int maxEdge = nodes[root]->getEdge(2)->getNumber();
    if(minEdge > maxEdge){
      int temp = minEdge;
      minEdge = maxEdge;
      maxEdge = temp;
    }
    if(edgeChosen>=minEdge)
      edgeChosen++;
    if(edgeChosen>=maxEdge)
      edgeChosen++;
    
    vector<int> indEdgesLeft(numEdges);
    int indCase;

    for(int i=0; i<numEdges; i++)
      indEdgesLeft[i] = i;
    if(edgeChosen==nodes[root]->getEdge(1)->getNumber() || edgeChosen==nodes[root]->getEdge(2)->getNumber()){
      //      cout << "indCase=1: edge from the root is chosen" << endl;
      indCase=1;
      indEdgesLeft[nodes[root]->getEdge(1)->getNumber()]=-1;
      indEdgesLeft[nodes[root]->getEdge(2)->getNumber()]=-1;
      accept = 0;
      acceptProb = 0;
      return;
    }
    else{
      //      cout << "indCase=2: edge not from the root is chosen" << endl;
      indCase=2;
      indEdgesLeft[edgeChosen]=-1;
      indEdgesLeft[edges[edgeChosen]->getNode(1)->getEdge(0)->getNumber()]=-1;
    }
    int numExtNodesLeft=numTaxa;
    edges[edgeChosen]->getNode(0)->travelSubTree(numExtNodesLeft, indEdgesLeft);
    int numEdgesLeft = numEdges-1;
    for(int i=numEdges-1; i>=0; i--)
      if(indEdgesLeft[i]==-1)
	indEdgesLeft[i] = indEdgesLeft[numEdgesLeft--];
    numEdgesLeft++;
    if(numEdgesLeft!= 2 * numExtNodesLeft - 2)
      cerr << "error: numEdgesLeft !=2*numExtNodesLeft-2" << endl;
    /*
    cout << "The Edge chosen is " << edgeChosen << endl;
    cout << "numExtNodesLeft= " << numExtNodesLeft << " numEdgesLeft= " << numEdgesLeft << endl;
    cout << "the edges left are: " << endl;
    for(int i=0; i<numEdgesLeft; i++)
      cout << indEdgesLeft[i] << endl;
    */
    if(numEdgesLeft==0){
      cerr << "error: numEdgesLeft=0" << endl;
      exit(1);
      accept = 0;
      acceptProb = 0;
      return;
    }
    int edgeAttached = indEdgesLeft[(int)(rand.runif() * numEdgesLeft)];

    setSavedValues5(edgeChosen, edgeAttached, indCase);

    double lengthAtt = edges[edgeAttached]->getLength();
    double tempLen = lengthAtt * rand.runif();
    int indSameEdge=0;
    if(indCase==1){
      // o
      //  \ /               o
      //   o              \/
      //   \/             o 
      //    o           \/
      //  e3\  /        o
      //atted\/        \/
      //      n2       o
      //       \   /  \/
      //      e1\ /e2 n3
      //         n1  /   
      //       e0 \ / chosen
      //          n0                
      if(savedNodes[1].getNumber()==savedNodes[2].getNumber()){
	edges[savedEdges[0].getNumber()]->setLength(tempLen);
	edges[savedEdges[2].getNumber()]->setNode(1, nodes[root]);
	edges[edgeAttached]->setNode(1,nodes[savedNodes[1].getNumber()]);
	edges[edgeAttached]->setLength(lengthAtt-tempLen);
	edges[edgeChosen]->setNode(1, nodes[savedNodes[1].getNumber()]);
	nodes[savedNodes[0].getNumber()]->setEdge(1,edges[savedEdges[0].getNumber()]);
	nodes[savedNodes[0].getNumber()]->setEdge(2,edges[savedEdges[2].getNumber()]);
	nodes[savedNodes[2].getNumber()]->setEdge(savedNodes[2].getEdge(1)->getNumber()==edgeAttached ? 2:1, edges[edgeChosen]);
      }
      else{
	edges[savedEdges[0].getNumber()]->setNode(1, nodes[savedNodes[2].getNumber()]);
	edges[savedEdges[0].getNumber()]->setLength(tempLen);
	edges[savedEdges[1].getNumber()]->setNode(1, nodes[root]);
	edges[savedEdges[2].getNumber()]->setNode(1, nodes[root]);
	edges[edgeAttached]->setNode(1,nodes[savedNodes[1].getNumber()]);
	edges[edgeAttached]->setLength(lengthAtt-tempLen);
	edges[edgeChosen]->setNode(1, nodes[savedNodes[1].getNumber()]);
	nodes[savedNodes[0].getNumber()]->setEdge(1,edges[savedEdges[1].getNumber()]);
	nodes[savedNodes[0].getNumber()]->setEdge(2,edges[savedEdges[2].getNumber()]);
	nodes[savedNodes[1].getNumber()]->setEdge(1,edges[edgeAttached]);
	nodes[savedNodes[1].getNumber()]->setEdge(2,edges[edgeChosen]);
	nodes[savedNodes[2].getNumber()]->setEdge(savedNodes[2].getEdge(1)->getNumber()==edgeAttached ? 1:2, edges[savedEdges[0].getNumber()]);
      }
    }
    else{
      // o
      //  \ /               o
      //   o              \/
      //    \/   chosen\  o 
      //     o          \/e0
      //     \/         n0
      //      o        \/e1   
      //      \        n1
      //     e2\     \/
      //atted   \    o
      //         n2 /   
      //         \ /
      //          o                
      if(edgeAttached==savedEdges[0].getNumber()){
	indSameEdge=1;
	lengthAtt = savedEdges[0].getLength()+savedEdges[1].getLength();
	tempLen = lengthAtt * rand.runif();
	edges[savedEdges[0].getNumber()]->setLength(lengthAtt-tempLen);
	edges[savedEdges[1].getNumber()]->setLength(tempLen);
      }
      else{
	edges[savedEdges[0].getNumber()]->setNode(1, nodes[savedNodes[1].getNumber()]);
	edges[savedEdges[0].getNumber()]->setLength(savedEdges[0].getLength()+savedEdges[1].getLength());
	edges[savedEdges[1].getNumber()]->setNode(1, edges[edgeAttached]->getNode(1));
	edges[savedEdges[1].getNumber()]->setLength(tempLen);
	edges[edgeAttached]->setNode(1,nodes[savedNodes[0].getNumber()]);
	edges[edgeAttached]->setLength(lengthAtt-tempLen);
	nodes[savedNodes[0].getNumber()]->setEdge(savedNodes[0].getEdge(1)->getNumber()==edgeChosen ? 2:1, edges[edgeAttached]);
	nodes[savedNodes[1].getNumber()]->setEdge(savedNodes[1].getEdge(1)->getNumber()==savedEdges[1].getNumber() ? 1:2, edges[savedEdges[0].getNumber()]);
	nodes[savedNodes[2].getNumber()]->setEdge(savedNodes[2].getEdge(1)->getNumber()==edgeAttached ? 1:2, edges[savedEdges[1].getNumber()]);
      }
    }
    logLikelihood = calculateNewLogLikelihood5(ad,rs, indCase);
    /*
    logLikelihoodX = calculateLogLikelihoodX(ad,rs);
    cout << "loglikelihood = " << logLikelihood << "         " << logLikelihoodX << endl;
    if(fabs(logLikelihood -logLikelihoodX) > 1.0e-10) {
      cerr << "Check Error: Likelihoods differ after proposal 4" << endl;
      cerr << "logLikelihood  = " << logLikelihood << endl;
      cerr << "logLikelihoodX = " << logLikelihoodX << endl;
      exit(1);
    }
    */
    if(indSameEdge==0)
      acceptProb = exp(logLikelihood - savedLogLikelihood) * savedEdges[2].getLength() / (savedEdges[1].getLength()+savedEdges[0].getLength()); // indCase=2;
    else
      acceptProb = exp(logLikelihood - savedLogLikelihood);        // indCase=2;
    if(rand.runif()<acceptProb)
      accept = 1;
    else{
      setBackValues5(indCase);
      logLikelihood = savedLogLikelihood;
    }
  }

  else{
    // move = 6
    // pick two edges and exchange them
    int edgeChosen1 = (int)((numEdges-2)*rand.runif());
    int edgeFromRoot1 = nodes[root]->getEdge(1)->getNumber();
    int edgeFromRoot2 = nodes[root]->getEdge(2)->getNumber();
    int minEdge, maxEdge;
    if(edgeFromRoot1 < edgeFromRoot2){
      minEdge = edgeFromRoot1;
      maxEdge = edgeFromRoot2;
    }
    else{
      minEdge = edgeFromRoot2;
      maxEdge = edgeFromRoot1;
    }
    if(edgeChosen1>=minEdge)
      edgeChosen1++;
    if(edgeChosen1>=maxEdge)
      edgeChosen1++;

    vector<int> edgeList(numEdges);
    vector<int> nodesToRoot1(numNodes);
    int length1=0;
    for(int i=0; i<numEdges; i++)
      edgeList[i] = i;
    edgeList[edgeChosen1]=-1;
    edgeList[edgeFromRoot1]=-1;
    edgeList[edgeFromRoot2]=-1;
    int temp=numTaxa;
    edges[edgeChosen1]->getNode(0)->travelSubTree(temp, edgeList);
    
    temp = edges[edgeChosen1]->getNode(1)->getNumber();
    edgeList[nodes[temp]->getEdge(1)->getNumber()==edgeChosen1 ? nodes[temp]->getEdge(2)->getNumber() : nodes[temp]->getEdge(1)->getNumber()]=-1;
    while(temp!=root){
      nodesToRoot1[length1++]=temp;
      Edge* tempEdge = nodes[temp]->getEdge(0);
      edgeList[tempEdge->getNumber()]=-1;
      temp = tempEdge->getNode(1)->getNumber();
    }
    int numEdgesLeft=numEdges-1;
    for(int i=numEdges-1; i>=0; i--)
      if(edgeList[i]==-1)
	edgeList[i]=edgeList[numEdgesLeft--];
    numEdgesLeft++;
    if(numEdgesLeft==0){
      acceptProb = 0;
      accept=0;
      return;
    }
    /*
    cout << "The edges avalaible are:" << endl;
    for(int i=0; i<numEdgesLeft; i++)
    cout << edgeList[i] << endl;*/
    int edgeChosen2 = edgeList[(int)(numEdgesLeft * rand.runif())];
    //    cout << "The edges chosen are: " << endl;
    //    cout << edgeChosen1 << setw(6) << edgeChosen2 << endl;

    setSavedValues6(edgeChosen1, edgeChosen2, nodesToRoot1, length1);
    Node *n1 = nodes[savedNodes[0].getNumber()];
    Node *n2 = nodes[savedNodes[1].getNumber()];
    edges[edgeChosen1]->setNode(1,n2);
    edges[edgeChosen2]->setNode(1,n1);
    n1->setEdge((n1->getEdge(1)->getNumber()==edgeChosen1 ? 1:2), edges[edgeChosen2]);
    n2->setEdge((n2->getEdge(1)->getNumber()==edgeChosen2 ? 1:2), edges[edgeChosen1]);

    logLikelihood = calculateNewLogLikelihood6(ad,rs);
    /*
    logLikelihoodX = calculateLogLikelihoodX(ad,rs);
    cout << "loglikelihood = " << logLikelihood << "         " << logLikelihoodX << endl;
    if(fabs(logLikelihood -logLikelihoodX) > 1.0e-10) {
      cerr << "Check Error: Likelihoods differ after proposal 4" << endl;
      cerr << "logLikelihood  = " << logLikelihood << endl;
      cerr << "logLikelihoodX = " << logLikelihoodX << endl;
      exit(1);
    }
    */
    acceptProb = exp(logLikelihood - savedLogLikelihood);

    if(rand.runif()<acceptProb)
      accept = 1;
    else{
      setBackValues6();
      logLikelihood = savedLogLikelihood;
    }

  }
}

void Tree::findNodesToRoot(Node* node, vector<int>& nodesList, int& length, int include, int root)
  // node, include, root are input parameters
  // nodesList, length are output parameters
  // find nodes from node (included or not) to root
{
  length=0;
  if(include==1)
    nodesList[length++]=node->getNumber();
  int temp=node->getNeighbor(0)->getNumber();
  while(temp!=root){
    nodesList[length++] = temp;
    temp = nodes[temp]->getNeighbor(0)->getNumber();
  }
}


void Node::travelSubTree(int& numExtNodesLeft, vector<int>& indEdgesLeft)                       // indicate the edges in the subtree with -1
{
  Node* node;
  if(isLeaf()){
    numExtNodesLeft--;
  }
  else{
    for(int i=1; i<3; i++){
      indEdgesLeft[getEdge(i)->getNumber()]=-1;
      node = getNeighbor(i);
      node->travelSubTree(numExtNodesLeft, indEdgesLeft);
    }
  }
}


void Tree::searchPath(Node* nodeS, Node* nodeE)
{
  indNodesInPath.resize(numNodes);
  indEdgesInPath.resize(numEdges);
  indNodesInPath[0]=nodeS->getNumber();
  int found=0;
  int m=1;

  Node* node=nodeS;
  vector<int> ancestors(numNodes);
  int mm=0;
  node->searchSibSubTree(found, nodeE, ancestors, mm, numNodes);
  while(!found){
    node=node->getNeighbor(0);
    indNodesInPath[m]=node->getNumber();
    m++;
    mm=0;
    node->searchSibSubTree(found,nodeE, ancestors, mm, numNodes);
  }
  for(int i=mm-1; i>=0; i--)
    indNodesInPath[m+(mm-1-i)]=ancestors[i];
  indNodesInPath[m+mm]=node->getNeighbor(0)->getNumber();
  numNodesInPath=m+mm+1;
  numEdgesInPath=numNodesInPath-1;
  for(int i=0; i<numEdgesInPath; i++)
    indEdgesInPath[i]=nodes[indNodesInPath[i]]->getEdge(0)->getNumber();
}


void Node::searchSibSubTree(int& found, Node* nodeE, vector<int>& ancestors, int& mm, int numNodes)
{
  Node* sib=getSib();
  //  cout << "sib is : " << sib->getNumber() << endl;
  //  cout << "numNodes= " << numNodes << endl;
  if(sib->isLeaf()){
    if(sib==nodeE){
      found=1;
      ancestors[0]=sib->getNumber();
      mm=1;
    }
  }
  else{
    ancestors[0]=sib->getNumber();
    vector<int> m(numNodes);
    m[sib->getNumber()]=mm; 
    sib->searchSubTree(found,nodeE,ancestors, mm, m);        
  }
}


void Node::searchSubTree(int& found, Node* nodeE, vector<int>& ancestors, int& mm, vector<int>& m)
{
  for(int i=1; i<3; i++){
    Node* n=this->getNeighbor(i);
    if(i==1) mm++;
    else mm=m[this->getNumber()]+1;
    m[n->getNumber()]=mm;
    ancestors[mm]=n->getNumber();/*
    //    cout << "node.number= " << this->getNumber() << "  " << n->getNumber() << endl;
    for(int j=0; j<=mm; j++)
      cout << "ancestors: " << ancestors[j] << endl;
				 */
    if(n->isLeaf()){ 
      if(n==nodeE){
	found=1;
	mm++;
	return;
      }
    }    
    else{
      n->searchSubTree(found, nodeE, ancestors, mm, m);
      if(found) return;
    }
  }
}

void Node::reOrderEdges(Node* parent)
{
  if(debug) {
    cerr << "reOrderEdges called on node[" << number << "] :";
    for(int i=0;i<3;i++)
      if(edges[i])
	cerr << "/" << edges[i]->getOtherNode(this)->getNumber();
      else
	cerr << "/";
    cerr << "/" << endl;
  }

  if(!leaf) {
    for(int i=1;i<3;i++)
      if(edges[i]->getOtherNode(this) == parent) {
	swapEdges(0,i);
	break;
      }
    for(int i=1;i<3;i++)
      edges[i]->getOtherNode(this)->reOrderEdges(this);
  }
  if(edges[0]->getNode(0) == parent)
    edges[0]->swapNodes();
}

void Node::reOrderEdgesRoot()
{
  for(int i=1;i<3;i++) // root has no edge 0 (no parent)
    edges[i]->getOtherNode(this)->reOrderEdges(this);
}

void Tree::reOrderEdges()
{
  nodes[root]->reOrderEdgesRoot();
}

void Tree::createRandomTree(const RunSettings& rs, Rand& rand)
{
  root = numNodes-1;
  nodes[0]->setEdge(0,edges[0]);
  nodes[1]->setEdge(0,edges[1]);
  nodes[root]->setEdge(1,edges[0]);
  nodes[root]->setEdge(2,edges[1]);
  edges[0]->setNodes(nodes[0], nodes[root]);
  edges[1]->setNodes(nodes[1], nodes[root]);

  int currentLeaf=2, currentInternal=numTaxa, currentExterEdge=2, currentInterEdge=numTaxa, currentEdge=2;

  for(int i=2; i<numTaxa; i++){

    // n1 -----> n2   ==>   n1 -----> n3 ------> n2
    //      e1                   e1    |     e2
    //                                 | e3
    //                                 V
    //                                n4
    int n=(int)(rand.runif()*currentEdge);
    if(n>=currentExterEdge)
      n = n-currentExterEdge+numTaxa;
    Edge *edge1=edges[n];
    Edge *edge2=edges[currentInterEdge++];
    Edge *edge3=edges[currentExterEdge++]; 
    currentEdge+=2;
    Node *node1=edge1->getNode(0);
    Node *node2=edge1->getNode(1);
    Node *node3=nodes[currentInternal++];
    Node *node4=nodes[currentLeaf++];

    edge1->setNode(1,node3);
    edge2->setNodes(node3,node2);
    if(node2->isLeaf())
      node2->setEdge(0,edge2);
    else
      for(int k=0;k<3;k++)
	if(node2->getEdge(k) == edge1) {
	  node2->setEdge(k,edge2);
	  break;
	}
    node3->setEdge(0,edge2);
    node3->setEdge(1,edge1);
    node3->setEdge(2,edge3);
    edge3->setNodes(node4,node3);
    node4->setEdge(0,edge3);
  }
  for(int i=0;i<numEdges;i++)
    edges[i]->setLength(-log(rand.runif()) * rs.getMu());
}


/*  
void Tree::createRandomTree(const RunSettings& rs,Rand& rand)
{
//First creat random unrooted tree, then put root in
  nodes[0]->setEdge(0,edges[0]);
  nodes[1]->setEdge(0,edges[0]);
  edges[0]->setNodes(nodes[0],nodes[1]);

  int currentLeaf=2,currentInternal=numTaxa,currentEdge=1;

  for(int i=2;i<numTaxa;i++) {

    // n1 -----> n2   ==>   n1 -----> n3 ------> n2
    //      e1                   e1    |     e2
    //                                 | e3
    //                                 V
    //                                n4
    int n=(int)(rand.runif()*currentEdge);
    Edge *edge1=edges[n];
    Edge *edge2=edges[currentEdge++];    
    Edge *edge3=edges[currentEdge++];    
    Node *node1=edge1->getNode(0);
    Node *node2=edge1->getNode(1);
    Node *node3=nodes[currentInternal++];
    Node *node4=nodes[currentLeaf++];

    edge1->setNode(1,node3);
    edge2->setNodes(node3,node2);
    if(node2->isLeaf())
      node2->setEdge(0,edge2);
    else
      for(int k=0;k<3;k++)
	if(node2->getEdge(k) == edge1) {
	  node2->setEdge(k,edge2);
	  break;
	}
    node3->setEdge(0,edge1);
    node3->setEdge(1,edge2);
    node3->setEdge(2,edge3);
    edge3->setNodes(node3,node4);
    node4->setEdge(0,edge3);
  }
  int n=(int)(rand.runif()*currentEdge);
  Edge *edge1=edges[n];
  Edge *edge2=edges[numEdges-1];
  Node *node1=edge1->getNode(0);
  Node *node2=edge1->getNode(1);
  Node *node3=nodes[numNodes-1];
  edge1->setNode(1,node3);
  edge2->setNodes(node3,node2);
  if(node2->isLeaf())
    node2->setEdge(0,edge2);
  else
    for(int k=0;k<3;k++)
      if(node2->getEdge(k)->getNumber() == edge2->getNumber()) {
	node2->setEdge(k,edge2);
	break;
      }
  node3->setEdge(1,edge1);
  node3->setEdge(2,edge2);
  root = numNodes-1;  
  reOrderEdges();
  for(int i=0;i<numEdges;i++)
    edges[i]->setLength(-rs.getMu()*log(rand.runif()));
}
*/


void Tree::connect2Nodes(Node *node1, Node *node2, double len, double len1)
{
  Edge *edge1, *edge2;
  edge1 = edges[node1->getNumber()];
  edge2 = edges[node2->getNumber()];
  node1->setEdge(0,edge1);
  node2->setEdge(0,edge2);
  root = numNodes-1;
  nodes[root]->setEdge(1,edge1);
  nodes[root]->setEdge(2,edge2);
  edge1->setLength(len1);
  edge2->setLength(len-len1);
  edge1->setNodes(node1,nodes[root]);
  edge2->setNodes(node2,nodes[root]);
}


Node *Tree::connect3Nodes(Node *node1, Node *node2, Node *node3, double len1, double len2)
{
  // node3 must be an internal node.
  Edge *edge1,*edge2;
  edge1 = edges[node1->getNumber()];
  edge2 = edges[node2->getNumber()];
  node1->setEdge(0,edge1);
  node2->setEdge(0,edge2);
  node3->setEdge(1,edge1);
  node3->setEdge(2,edge2);
  edge1->setNodes(node1,node3);
  edge1->setLength(len1);
  edge2->setNodes(node2,node3);
  edge2->setLength(len2);
  return node3;
}


void Tree::createGoodTree(const AFLPData& ad, Rand& rand)
{
  vector<vector<double> > dist(numTaxa);
  for(int i=0;i<numTaxa;i++)
    dist[i].resize(numTaxa);
  ad.makeDistMatrix(dist);

  vector<Node*> available(numTaxa);
  int numAvailable=numTaxa, currentInternal=numTaxa;
  for(int i=0;i<numTaxa;i++)
    available[i] = nodes[i];

  for(int k=0;k<numTaxa-2;k++) {
    int n1=0,n2=1;
    vector<double> u(numAvailable);
    for(int i=0; i<numAvailable; i++){
      u[i] = 0;
      for(int j=0; j<numAvailable; j++)
	if(j!=i)
	  u[i] = u[i] + dist[i][j];
      //      u[i] = u[i]/(double)(numAvailable-2);
      u[i] = u[i]/(double)(numTaxa-2);
    }
    for(int i=0;i<numAvailable-1;i++)
      for(int j=i+1;j<numAvailable;j++)
	if((dist[i][j]-u[i]-u[j]) < (dist[n1][n2]-u[n1]-u[n2])){
	  n1 = i;
	  n2 = j;
	}
    double len1 = (dist[n1][n2] + u[n1] - u[n2])/2;
    double len2 = (dist[n1][n2] + u[n2] - u[n1])/2;
    if(len1<0){
      len1=0;
      len2=dist[n1][n2];
    }
    else if(len2<0){
      len1=dist[n1][n2];
      len2=0;
    }
    available[n1] = connect3Nodes(available[n1],available[n2],nodes[currentInternal++], len1, len2);
    available[n2] = available[numAvailable-1];

    for(int i=0;i<numAvailable-1;i++) { 
      dist[n1][i] = dist[i][n1] = (dist[n1][i]+dist[n2][i]-dist[n1][n2])/2;
      dist[n2][i] = dist[i][n2] = dist[numAvailable-1][i];
    }
    numAvailable--;
  }
  connect2Nodes(available[0],available[1], dist[0][1], dist[0][1]*rand.runif());   
}


void Tree::printNodesEdges() const
{
  if(debug) {
    cerr << "Nodes: " << endl;
    for(int k=0;k<numNodes;k++) {
      for(int i=0;i<3;i++) {
	Edge *edge = nodes[k]->getEdge(i);
	if(edge)
	  cerr << "/" << edge->getOtherNode(nodes[k])->getNumber();
	else
	  cerr << "/";
      }
      cerr << "/" << endl;
    }
    cerr << "Edges: " << endl;
    for(int k=0;k<numEdges;k++) {
      for(int i=0;i<2;i++)
	cerr << "/" << edges[k]->getNode(i)->getNumber();
      cerr << "/" << endl;
    }
  }
}

void Tree::check(int ind[6][6]) const
{
  for(int i=0; i<6; i++)
    for(int j=i+1; j<6; j++)
      if(nodes[i]->getNeighbor(0) == nodes[j]->getNeighbor(0)){
	ind[i][j]++;
	ind[j][i]++;
      }
}


void Tree::print(ostream& f) const
{
  f << setiosflags(ios::showpoint | ios::fixed);

  f << "  #  lc  rc  pt e0  e1  e2  len" << endl;
  for(int i=0;i<numTaxa;i++)
    f << setw(3) << nodes[i]->getNumber() << setw(4) << -1 << setw(4) << -1
      << setw(4) << nodes[i]->getEdge(0)->getOtherNode(nodes[i])->getNumber()
      //      << setw(4) << nodes[i]->getEdge(0)->getNode(1)->getNumber()
      << setw(4) << nodes[i]->getEdge(0)->getNumber()
      << setw(4) << -1 << setw(4) << -1
      << setw(6) << setprecision(3) << nodes[i]->getEdge(0)->getLength() << endl;
  for(int i=numTaxa;i<numNodes;i++)
    if(i!=root)
      f << setw(3) << nodes[i]->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(0)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(0)->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getNumber()
	<< setw(6) << setprecision(3) << nodes[i]->getEdge(0)->getLength() << endl;
    else
      f << setw(3) << nodes[i]->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << -1
	<< setw(4) << -1
	<< setw(4) << nodes[i]->getEdge(1)->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getNumber()
	<< setw(6) << setprecision(3) << 0.0 << endl;
}
