#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <sstream>

using namespace std;

int minimum(int a, int b) { return (a < b ? a : b); }

class Node;

class Edge {
public:
  Edge () {}
  Edge(int n) :
    number(n)
  {}
  int getNumber() { return number; }
  Node* getOtherNode(Node *n) { return (nodes[0] == n ? nodes[1] : nodes[0]); }
  void setNode(int i,Node *n) { nodes[i]=n; }
  void setNodes(Node *m,Node *n) { nodes[0]=m; nodes[1]=n; }
  Node* getNode(int n) { return nodes[n]; }
  void swapNodes() {
    Node *temp=nodes[0];
    nodes[0] = nodes[1];
    nodes[1] = temp;
  }
  void print(ostream&);
private:
  int number;
  Node* nodes[2];
};

class Node {
 public:
  Node() {}
  Node(int n,int l) : number(n), leaf(l) {}
  int getNumber() const { return number; }
  int isLeaf() { return leaf; }
  Edge* getEdge(int n) { return edges[n]; }
  void setEdge(int n,Edge *e) { edges[n] = e; }
  void swapEdges(int x,int y) {
    Edge *temp = edges[x];
    edges[x] = edges[y];
    edges[y] = temp;
  }
  Node* getNeighbor(int n) { return edges[n]->getOtherNode(this); }
  Node* getSib(){
    for(int i=1; i<3; i++)
      if(edges[0]->getOtherNode(this)->getNeighbor(i) == this){
	return edges[0]->getOtherNode(this)->getNeighbor(3-i);
	break;
      }
  }
  void print(ostream&);
  void reOrderEdges(Node*);
  void reOrderEdgesRoot();
 private:
  int number;
  int leaf;
  Edge *edges[3];
};

class Tree {
public:
  Tree() {};
  Tree(string);
  ~Tree() {
    for(int i=0;i<numNodes;i++)
      delete nodes[i];
    for(int i=0;i<numEdges;i++)
      delete edges[i];
  }
  void setNodesAndEdges(int,double*);
  int getNumLeaves() { return numLeaves; }
  int getNumNodes() { return numNodes; }
  int getNumEdges() { return numEdges; }
  int getRoot() { return root; }
  Node* getNode(int n) { return nodes[n]; }
  Edge* getEdge(int n) { return edges[n]; }
  void findSmallest(vector<int>&,Node*) const;
  void printIntTop(ostream&,vector<int>&, Node*) const;
  void printTopology(ostream&) const;
  void print(ostream&) const;
  void reOrderEdges();
  Node* readTreeInt(istringstream&,int&,int&);
  Node* connect3Nodes(Node*,Node*,Node*,int&);
  void reroot(int);
private:
  int numLeaves;
  int numNodes;
  int numEdges;
  vector<Node*> nodes;
  vector<Edge*> edges;
  int root;
  int rooted;
};

Tree::Tree(string top) {
  int numLeft=0,numRight=0,numComma=0;
  for(int i=0;i<top.length();i++) {
    switch(top[i]) {
    case '(' :
      numLeft++;
      break;
    case ')' :
      numRight++;
      break;
    case ',' :
      numComma++;
      break;
    }
  }
  if(numLeft != numRight)
    cerr << "Error: Unequal numbers of left and right parentheses." << endl;
  numLeaves = numComma+1;
  rooted = (numLeft==numComma ? 1 : 0);
  if(!rooted) {
    cerr << "Error: Program does not read in unrooted trees." << endl;
    exit(1);
  }

  numNodes = (rooted ? 2*numLeaves - 1 : 2*numLeaves - 2);
  numEdges = numNodes - 1;
  nodes.resize(numNodes);
  edges.resize(numEdges);
  for(int i=0;i<numLeaves;i++)
    nodes[i] = new Node(i,1);
  for(int i=numLeaves;i<numNodes;i++)
    nodes[i] = new Node(i,0);
  for(int i=0;i<numEdges;i++)
    edges[i] = new Edge(i);

  int currentInternal=numLeaves,currentInternalEdge=numLeaves;
  istringstream s(top);
  Node *node1,*node2;
  char ch;
  int n;
  s >> ch;
  if(ch!='(') {
    s.putback(ch);
    s >> n;
    if(numNodes>1) {
      cerr << "Error: Tree with more than one leaf should begin with `)'" << endl;
      exit(1);
    }
  }
  else {
    node1 = readTreeInt(s,currentInternal,currentInternalEdge);
    s >> ch;
    if(ch!=',') {
      cerr << "Error: Cannot read in tree, expected ',', found " << ch << endl; 
     exit(1);
    }
    node2 = readTreeInt(s,currentInternal,currentInternalEdge);
    s >> ch;
    if(ch!=')') {
      cerr << "Cannot read in tree, expected ')', found " << ch << endl;
      exit(1);
    }
    connect3Nodes(node1,node2,nodes[currentInternal],currentInternalEdge);
    root = currentInternal;
  }
}

Node* Tree::readTreeInt(istringstream& s, int& currentInternal, int &currentInternalEdge)
{
  char ch;
  int n;
  s >> ch;
  if(ch!='(') {
    s.putback(ch);
    s >> n;
    if(n>numLeaves) {
      cerr << "Error: Taxon number (" << n << ") is larger than the number of leaves (" << numLeaves << ")" << endl;
      exit(1);
    }
    return nodes[n-1];
  }
  else {
    Node *node1,*node2;
    node1 = readTreeInt(s,currentInternal,currentInternalEdge);
    s >> ch;
    if(ch!=',') {
      cerr << "Error: Cannot read in tree, expected ',', found " << ch << endl;
      exit(1);
    }
    node2 = readTreeInt(s,currentInternal,currentInternalEdge);
    s >> ch;
    if(ch!=')') {
      cerr << "Error: Cannot read in tree, expected ')', found " << ch << endl;
      exit(1);
    }
    return connect3Nodes(node1,node2,nodes[currentInternal++],currentInternalEdge);
  }
}

Node *Tree::connect3Nodes(Node *node1, Node *node2, Node *node3, int &currentInternalEdge)
{
  // node3 must be an internal node.
  Edge *edge1,*edge2;
  edge1 = (node1->isLeaf() ? edges[node1->getNumber()] : edges[currentInternalEdge++]);
  edge2 = (node2->isLeaf() ? edges[node2->getNumber()] : edges[currentInternalEdge++]);
  node1->setEdge(0,edge1);
  node2->setEdge(0,edge2);
  node3->setEdge(1,edge1);
  node3->setEdge(2,edge2);
  edge1->setNodes(node1,node3);
  edge2->setNodes(node2,node3);
  return node3;
}

void Tree::findSmallest(vector<int> &smallest, Node* node) const
{
  int num = node->getNumber();
  if(node->isLeaf())
    smallest[num] = num;
  else {
    Node *node1 = node->getNeighbor(1);
    findSmallest(smallest,node1);
    int n1 = smallest[node1->getNumber()];
    Node *node2 = node->getNeighbor(2);
    findSmallest(smallest,node2);
    int n2 = smallest[node2->getNumber()];
    smallest[num] = minimum(n1,n2);
  }
}

void Tree::printIntTop(ostream& f, vector<int> &smallest, Node* node) const
{
  if(node->isLeaf()) {
    f << node->getNumber()+1;
  }
  else {
    Node *node1 = node->getNeighbor(1);
    Node *node2 = node->getNeighbor(2);
    if(smallest[node1->getNumber()] < smallest[node2->getNumber()]) {
      f << "(";
      printIntTop(f,smallest,node1);
      f << ",";
      printIntTop(f,smallest,node2);
      f << ")";
    }
    else {
      f << "(";
      printIntTop(f,smallest,node2);
      f << ",";
      printIntTop(f,smallest,node1);
      f << ")";
    }
  }
}

void Tree::printTopology(ostream& f) const
{
  vector<int> smallest(numNodes);
  findSmallest(smallest,nodes[root]);
  printIntTop(f,smallest,nodes[root]);
  f << ";" << endl;
}

void Tree::reroot(int outGroup) {
  Node* node1 = nodes[outGroup]->getNeighbor(0);
  if(node1->getNumber() == root)
    return;
  int c=1;
  while(node1->getNeighbor(c)->getNumber() != outGroup)
    c++;
  Edge* edge = node1->getEdge(0);
  Node* node2 = node1;
  Node* node3 = node2->getNeighbor(0);
  while(node3->getNumber() != root) {
    node2->setEdge(c,edge);
    int d = 1;
    while(node3->getNeighbor(d) != node2)
      d++;
    node3->setEdge(d,edge);
    c = d;
    edge->swapNodes();

    edge = node3->getEdge(0);
    node2 = node3;
    node3 = edge->getNode(1);
  }
  node1->setEdge(0,edge);
  node2->getSib()->getEdge(0)->setNode(1,node2);
  node2->setEdge(c,node2->getSib()->getEdge(0));
  node3->setEdge(3-c,nodes[outGroup]->getEdge(0));
  edge->setNode(0,node1);
}

void Tree::print(ostream& f) const
{
  f << setiosflags(ios::showpoint | ios::fixed);

  f << "  #  lc  rc  pt e0  e1  e2  len" << endl;
  for(int i=0;i<numLeaves;i++)
    f << setw(3) << nodes[i]->getNumber() << setw(4) << -1 << setw(4) << -1
      << setw(4) << nodes[i]->getEdge(0)->getOtherNode(nodes[i])->getNumber()
      //      << setw(4) << nodes[i]->getEdge(0)->getNode(1)->getNumber()
      << setw(4) << nodes[i]->getEdge(0)->getNumber()
      << setw(4) << -1 << setw(4) << -1
      << setw(6) << setprecision(3) << 1 << endl;
  for(int i=numLeaves;i<numNodes;i++)
    if(i!=root)
      f << setw(3) << nodes[i]->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(0)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(0)->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getNumber()
	<< setw(6) << setprecision(3) << 1 << endl;
    else
      f << setw(3) << nodes[i]->getNumber()
	<< setw(4) << nodes[i]->getEdge(1)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getOtherNode(nodes[i])->getNumber()
	<< setw(4) << -1
	<< setw(4) << -1
	<< setw(4) << nodes[i]->getEdge(1)->getNumber()
	<< setw(4) << nodes[i]->getEdge(2)->getNumber()
	<< setw(6) << setprecision(3) << 0.0 << endl;
}

int main(void) {
  string top;
  while(getline(cin,top)) {
    Tree t(top);
    t.reroot(0);
    t.printTopology(cout);
  }
  return 0;
}
  
